/*
 * Copyright (C) 2003-2007 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "base/Log.h"
#include "winmaincpp.h"
#include "utils.h"
#include "vObjectConverter.h"
#include "vocl/VConverter.h"
#include "outlook/utils.h"
#include "outlook/ClientItem.h"
#include "outlook/ClientAppException.h"
#include "outlook/ClientException.h"
#include "SIFFields.h"

#include <string>

using namespace std;


/**
 * ClientItem object -> SyncItem object (client to server).
 * SyncItem data is the string generated from all ClientItem properties.
 * Data is converted into SIF or other mime types, based on 'dataType' parameter.
 *
 * @param cItem    : [INPUT] pointer to ClientItem
 * @param dataType : the mime type of data we want into the SyncItem (SIF/VCard/...)
 * @return         : the (new allocated) SyncItem object
 *                   Returned pointer MUST be freed by the caller
 */
SyncItem* convertToSyncItem(ClientItem* cItem, const char* dataType) {

    if (!cItem) {
        return NULL;
    }

    wstring id   = cItem->getID();
    wstring type = cItem->getType();
    if (id == EMPTY_WSTRING) {
        return NULL;
    }

    SyncItem* sItem = new SyncItem(id.c_str());
    char* data = NULL;


    //
    // ----------- SIF data type -----------
    //
    if (isSIF(dataType)) {

        WCHAR** sifFields = getProperSifArray(type);
        wstring sif, propertyValue;

        //
        // Generate SIF string from ClientItem.
        //
        sif  = L"<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n<";
        sif += type;
        sif += L">\n";

        wstring path = cItem->getParentPath();

        // Replace "\\Personal Folders\Contacts" -> "DEFAULT_FOLDER"
        ClientApplication* outlook = ClientApplication::getInstance();
        ClientFolder* folder = outlook->getDefaultFolder(type);
        if (folder) {
            wstring defFolderPath = folder->getPath();
            replaceAll(defFolderPath, DEFAULT_FOLDER, path);
        }

        replaceAll(L"&", L"&amp;", path);
        replaceAll(L"<", L"&lt;",  path);
        replaceAll(L">", L"&gt;",  path);

        sif += L"<Folder>";
        sif += path;
        sif += L"</Folder>\n";

        for (int i=0; sifFields[i]; i++) {
            
            propertyValue = cItem->getProperty(sifFields[i]);
            
            if (propertyValue != EMPTY_WSTRING) {
                replaceAll(L"&", L"&amp;", propertyValue);
                replaceAll(L"<", L"&lt;",  propertyValue);
                replaceAll(L">", L"&gt;",  propertyValue);

                sif += L"<";
                sif += sifFields[i];
                sif += L">";
                sif += propertyValue;
                sif += L"</";
                sif += sifFields[i];
                sif += L">\n";
            }
            else {
                sif += L"<";
                sif += sifFields[i];
                sif += L"/>\n";
            }
        }

        // Manage appointment exceptions.
        if (cItem->getType() == APPOINTMENT) {
            sif += formatAppExceptions((ClientAppointment*)cItem);
        }

        sif += L"</";
        sif += type;
        sif += L">\n";

        data = toMultibyte(sif.c_str());
    }

    //
    // ----------- Other data type (VCard/VCal...) -----------
    //
    else {
        //
        // TBD: check if is an accepted mime type
        //

        VObject* vobj = NULL;

        if (cItem->getType() == CONTACT) {
            vobj = client2VObjectContact((ClientContact*)cItem);
        }
        else if (cItem->getType() == APPOINTMENT) {
            vobj = client2VObjectAppointment((ClientAppointment*)cItem);
        }
        // *** TBD ***
        //else if (cItem->getType() == TASK) {
        //    vobj = client2VObjectTask((ClientTask*)cItem);
        //}

        if (vobj) {
            WCHAR* wdata = vobj->toString();
            data = toMultibyte(wdata);

            if (wdata) {
                delete [] wdata;
            }
        }
    }


    //
    // Set SyncItem data.
    //
    if (data) {
        // Just replace chars < 0x20 (excluding 0x0a, 0x0d, 0x09).
        checkIllegalXMLChars(data);

        long dataSize = strlen(data);
        sItem->setData(data, dataSize);
        delete [] data;
    }

    return sItem;
}


/**
 * Check for illegal XML chars inside 'data'.
 * Chars < 0x20 are converted to 0x21 (space) - preserved '\t' '\n' '\r'.
 *
 * @param  data   [IN/OUT] the buffer to analyze
 * @return        1 if some char is converted, 0 otherwise
 */
int checkIllegalXMLChars(char* data) {
    
    if (!data) return 0;
    int len = strlen(data);

    int ret = 0;
    for (int i=0; i<len; i++) {
        if (data[i] < 0x20) {
            if ( (data[i] != 0x09) &&       // '\t'
                 (data[i] != 0x0a) &&       // '\n'
                 (data[i] != 0x0d) ) {      // '\r'
                data[i] = 0x20;             // ' '
                ret = 1;
            }
        }
    }
    return ret;
}




/**
 * Format appointment exceptions (client to server).
 *
 * @param cApp : the ClientAppointment pointer to retrieve data
 * @return     : a SIF string formatted like this:
 *
 *<Exceptions>
 *   <ExcludeDate>...</ExcludeDate>         // each of these is an exception (deleted occurrence)
 *   <ExcludeDate>...</ExcludeDate>
 *   <ExcludeDate>...</ExcludeDate>
 *   <IncludeDate/>                         // Outlook does not support added occurrences: we MUST send this empty.
 * </Exceptions>
 */ 
wstring formatAppExceptions(ClientAppointment* cApp) {

    if (!cApp) return EMPTY_WSTRING;
    ClientRecurrence* cRec = cApp->getRecPattern();
    if (!cRec) {
        return L"<Exceptions/>\n";
    }


    // Correct duplicated exceptions from Outlook...
    cRec->removeDuplicatedExceptions();

    wstring sif = EMPTY_WSTRING;
    int numExceptions = cRec->getExceptionsCount();

    //
    // Format all exceptions:
    //
    if (numExceptions > 0) {
        ClientAppException* exc = NULL;
        wstring propertyValue;

        // These are necessary to correctly manage "OriginalDate" property!
        BOOL isAllDay = cApp->isAllDayEvent();
        wstring start = cApp->getProperty(L"Start");

        sif += L"<Exceptions>\n";
        
        // Cycle through Exceptions:
        for (int i=0; i<numExceptions; i++) {

            exc = cRec->getException(i);

            // NOTE: Only deleted occurrences should be retrieved here!
            // (see 'normalizeException()' function)
            sif += L"<ExcludeDate>";
            sif += exc->formatOriginalDate(isAllDay, start);
            sif += L"</ExcludeDate>\n";
        }

        // ALWAYS SEND THIS FIELD EMPTY, to ensure that any added occurrence
        // will be cancelled on the other side.
        sif += L"<IncludeDate/>\n";

        sif += L"</Exceptions>\n";
    }

    // No exceptions: add <Exceptions/> to ensure that any exception
    // will be cancelled on the other side.
    else {
        sif = L"<Exceptions/>\n";
    }

    return sif;
}




/**
 * data string (SIF/VCard/...) -> ClientItem (server to client).
 * Fill the passed ClientItem object with all properties from 'data' string.
 * Data is parsed from SIF or other mime types, based on 'dataType' parameter.
 *
 * @param data     : the input string to read from
 * @param cItem    : [IN-OUT] the ClientItem object to fill
 * @param itemType : the type of item (contact/task/...)
 * @param dataType : the mime type of data sent by server
 * @return         : 0 if no errors
 */
int fillClientItem(const wstring& data, ClientItem* cItem, const wstring& itemType, const WCHAR* dataType) {


    // If appointment, we clear the recurrence pattern (all fields are always sent). 
    // Also appointment exceptions are removed here (if any).
    if (itemType == APPOINTMENT) {
        ((ClientAppointment*)cItem)->clearRecPattern();
    }


    //
    // ----------- SIF data type -----------
    //
    if (isSIF(dataType)) {
        LOG.debug(DBG_PARSING_SIF_DATA, itemType.c_str());

        wstring propertyValue;
        WCHAR** sifFields = getProperSifArray(itemType);
       
        // Check if <itemType> tag is present...
        wstring::size_type pos = 0;
        wstring itemTypeTag = L"<";
        itemTypeTag += itemType;
        itemTypeTag += L">";
        pos = data.find(itemTypeTag, 0);
        if (pos == wstring::npos) {
            LOG.error(ERR_PARSE_SIF_TAG_NOT_FOUND, itemTypeTag.c_str());
            return 1;
        }

        for (int i=0; sifFields[i]; i++) {
            // Set only properties found!
            if (!getElementContent(data, sifFields[i], propertyValue, 0)) {

                replaceAll(L"&lt;",  L"<", propertyValue);
                replaceAll(L"&gt;",  L">", propertyValue);
                replaceAll(L"&amp;", L"&", propertyValue);

                cItem->setProperty(sifFields[i], propertyValue);
            }
        }

        // Manage appointment exceptions.
        if (itemType == APPOINTMENT) {
            parseAppExceptions(data, cItem);
        }
    }

    //
    // ----------- Other data type (VCard/VCal...) -----------
    //
    else {
        LOG.debug(DBG_PARSING_VOBJ_DATA, itemType.c_str(), dataType);

        VObject* vobj = VConverter::parse(data.c_str());
        if (!vobj) {
            sprintf(lastErrorMsg, ERR_ITEM_VOBJ_PARSE, dataType);
            LOG.error(lastErrorMsg);
            return 1;
        }

        // Check if VObject type and version are the correct ones.
        if ( !checkVobjectTypeAndVersion(vobj, cItem->getType()) ) {
            if (vobj) delete vobj;
            return 1;
        }

        if (cItem->getType() == CONTACT) {
            vObject2ClientContact(vobj, (ClientContact*)cItem);
        }
        else if (cItem->getType() == APPOINTMENT) {
            vObject2ClientAppointment(vobj, (ClientAppointment*)cItem);
        }
        // *** TBD ***
        //else if (cItem->getType() == TASK) {
        //    vObject2ClientTask(vobj, (ClientTask*)cItem);
        //}

        if (vobj) {
            delete vobj;
        }
    }

    return 0;
}



/**
 * Parse appointment exceptions from SIF data (server to client).
 * Lists of exceptions are created and passed to 'setRecurrenceExceptions()' function.
 * 
 * @param sifString : the SIF string to read from
 * @param cItem     : the ClientItem object to modify (it's an appointment)
 * @return          : 0 if exceptions saved with no errors
 *                    1 if nothing done
 *                   -1 if errors
 */
int parseAppExceptions(const wstring& sifString, ClientItem* cItem) {

    wstring exString = EMPTY_WSTRING;

    if (getElementContent(sifString, L"Exceptions", exString)) {
        // Not found: nothing to do.
        return 1;
    }

    list<wstring> excludeDates, includeDates;
    excludeDates.clear();
    includeDates.clear();


    // <Exceptions> not empty -> parse all exceptions
    if (exString.size() > 0) {
        wstring excludeDate = EMPTY_WSTRING;
        wstring includeDate = EMPTY_WSTRING;
        wstring::size_type pos, start, end;

        // Parse all <ExcludeDate>
        pos = 0;
        while ( !getElementContent(exString, L"ExcludeDate", excludeDate, pos, start, end) ) {
            pos = end;
            if (excludeDate.size() > 0) {
                excludeDates.push_back(excludeDate);
            }
        }

        // Parse all <IncludeDate>
        pos = start = end = 0;
        while ( !getElementContent(exString, L"IncludeDate", includeDate, pos, start, end) ) {
            pos = end;
            if (includeDate.size() > 0) {
                includeDates.push_back(includeDate);
            }
        }
    }


    // Add exceptions to Outlook: new events could be created in Outlook
    // (added occurrences are not supperted by Outlook)
    return setRecurrenceExceptions(cItem, excludeDates, includeDates);
}





/**
 * Utility to return the right pointer to the static WCHAR** array 
 * of SIF fields, given the 'type' itemType.
 */
WCHAR** getProperSifArray(const wstring& type) {

    if (type == CONTACT) {
        return contactFields;
    }
    else if (type == APPOINTMENT) {
        return appointmentFields;
        }
    else if (type == TASK) {
        return taskFields;
        }
    else if (type == NOTE) {
        return noteFields;
    }
    else {
        // error...
        return NULL;
    }
}





/**
 * Normalize all appointment exceptions for passed item (client to server).
 * This is done before sending items to server. After normalization we will have
 * only exceptions that are 'deleted occurrences', and new unlinked appointments for
 * other exceptions. New appointments are also added to the 'allItems' list.
 * In case of errors, 'lastErrorMsg' is set and '1' is returned.
 *
 * @param allItems : [IN-OUT] the list of all appointments to scan
 * @return         : 0 if no errors
 */
int normalizeExceptions(ClientItem* cItem, itemKeyList& allItems) {

    ClientAppointment* cApp = (ClientAppointment*)cItem;
    if (!cApp) {
        // Maybe is not an appointment.
        return 1;
    }

    ClientRecurrence* cRec = cApp->getRecPattern();
    if (!cRec) {
        // Not recurring -> no exceptions
        return 0;
    }

    int exCount = cRec->getExceptionsCount();
    if (exCount == 0) {
        // No exceptions
        return 0;
    }


    //
    // Scan all exceptions
    //
    for (int i=0; i<exCount; i++) {
        ClientAppException* cEx = cRec->getException(i);
        if (cEx->getDeleted()) {
            // Deleted exceptions are OK.
            continue;
        }
        else {
            // Modified exceptions must be converted.
            // --------------------------------------
            LOG.debug(DBG_NORMALIZING_EXCEPTION, getSafeItemName(cItem).c_str(), cEx->formatOriginalDate().c_str());

            //
            // 1. Create new (unlinked) single event.
            //
            // Copy from main appointment (all original props are copied).
            ClientItem* cItemNew = cItem->copyItem();

            if (cItemNew) {
                // MUST clear the rec pattern...
                cItemNew->setProperty(L"IsRecurring", L"0");
                ((ClientAppointment*)cItemNew)->clearRecPattern();

                // Set exception properties on the new appointment.
                wstring propertyValue = EMPTY_WSTRING;
                for (int j=0; exAppointmentFields[j]; j++) {
                    propertyValue = cEx->getAppProperty(exAppointmentFields[j]);
                    cItemNew->setProperty(exAppointmentFields[j], propertyValue);
                }

                // Save the new appointment and add its LUID to the new items list.
                if (!cItemNew->saveItem()) {
                    allItems.push_back(cItemNew->getID());
                }
                else {
                    sprintf(lastErrorMsg, ERR_ITEM_SAVE, APPOINTMENT, getSafeItemName(cItemNew).c_str(), cItemNew->getParentPath().c_str());
                    goto error;
                }
                delete cItemNew;
                cItemNew = NULL;
            }
            else {
                sprintf(lastErrorMsg, ERR_ITEM_CREATE, APPOINTMENT, cItemNew->getParentPath().c_str());
                goto error;
            }

            //
            // 2. Modify exception -> convert into a 'occurrence deleted'
            //    Must also save the recurring appointment (it's modified)
            //
            cEx->setDeleted(TRUE);
            if (cItem->saveItem()) {
                sprintf(lastErrorMsg, ERR_ITEM_SAVE, L"recurring appointment", getSafeItemName(cItem).c_str(), cItem->getParentPath().c_str());
                goto error;
            }

            //
            // 3. Delete also ALL (existing) occurrences between originalDate and StartDate!
            //    (create an exception 'occurrence deleted' for each occurrence found).
            //
            DATE startDate  = NULL;
            systemTimeDateToDouble(cEx->getStart(), &startDate);
            DATE originalDate = cEx->getOriginalDate();
            int numExCreated = deleteOccurrencesInInterval(startDate, originalDate, cRec);

            //
            // 4. Save the recurring appointment (if it's modified)
            //
            if (numExCreated > 0) {
                if (cItem->saveItem()) {
                    sprintf(lastErrorMsg, ERR_ITEM_SAVE, L"recurring appointment", getSafeItemName(cItem).c_str(), cItem->getParentPath().c_str());
                    goto error;
                }
            }
        }
    } // end: for (int i=0; i<exCount; i++)

    return 0;

error:
    LOG.error(lastErrorMsg);
    return 1;
}





/**
 * Deletes all existing occurrences in the interval [startDate - originalDate[ , 
 * from ClientRecurrence object 'cRec'.
 * When an occurrence is found in the given interval, a proper exception
 * is added to 'cRec' (a 'deleted occurrence' exception).
 * 
 * @param startDate    : the start date (included in interval)
 * @param originalDate : the original date (excluded from interval)
 * @param cRec         : the ClientRecurrence pointer to work on
 * @return             : the number of occurrences deleted 
 *                       ( = number of exceptions created in cRec)
 */
int deleteOccurrencesInInterval(const DATE startDate, const DATE originalDate, ClientRecurrence* cRec) {

    int exCreated = 0;
    int originalDay = (int)originalDate;
    int startDay    = (int)startDate;
    DATE date1, date2;
    _AppointmentItemPtr pOcc = NULL;

    // Same day: nothing to do
    if (originalDay == startDay) {
        return 0;
    }

    // Get date interval [date1 - date2] of occurrences to delete.
    if (originalDay < startDay) {
        date1 = originalDate + 1;
        date2 = startDate;
    }
    else {
        date1 = startDate;
        date2 = originalDate - 1;
    }

    // Delete all occurrences found between date1 and date2.
    // (create exceptions as 'occurrence deleted').
    for (DATE i=date1; i<=date2; i++) {

        if (cRec->getOccurrence(i)) {
            ClientAppException* cExNew = new ClientAppException();

            cExNew->setOriginalDate(i);
            cExNew->setDeleted(TRUE);
            cRec->addException(cExNew);

            delete cExNew; cExNew = NULL;
            exCreated ++;
        }
    }
    return exCreated;
}





/**
 * Sets the appointment exceptions for the item passed (server to client).
 * Exceptions dates are passed with 2 lists 'escludeDates' and 'includeDates'.
 * Each exception found is set into a ClientAppException object. Each ClientRecurrence
 * object can have a list of ClientAppExceptions.
 * All exceptions are finally saved to Outlook when calling the method 'save()' of
 * the corrispondent ClientAppointment object.
 *
 * - Deleted occurrences are normally added to Client
 * - Added occurrences are not supported by Outlook 
 *   -> a new (unlinked event) is created
 *   -> the item is modified itself, must be added next sync to the MOD items list
 * 
 * @param cItem        : the ClientItem object to modify (it's an appointment)
 * @param excludeDates : a list of exclude-dates (occurrences to delete)
 * @param includeDates : a list of include-dates (occurrences to add)
 * @return             : 0 if exceptions saved with no errors
 *                       1 if not recurring (nothing done)
 *                      -1 if errors
 */
int setRecurrenceExceptions(ClientItem* cItem, list<wstring> &excludeDates, list<wstring> &includeDates) {

    if (!cItem) return -1;
    ClientAppointment*  cApp = (ClientAppointment*)cItem;
    if (!cApp)  return -1;
    ClientRecurrence*   cRec = cApp->getRecPattern();

    // Not recurring: nothing to do.
    if (!cRec) {
        if (excludeDates.size() || includeDates.size()) {
            LOG.error(ERR_PARSE_EXC_NOREC);
            return -1;
        }
        return 1;
    }

    // No data: reset exceptions.
    if (!excludeDates.size() && !includeDates.size()) {
        cRec->resetExceptions();
        return 0;
    }


    ClientAppException* cOcc = NULL;
    wstring excludeDate      = EMPTY_WSTRING;
    wstring includeDate      = EMPTY_WSTRING;
    list<wstring>::iterator  exIterator;
    list<wstring>::iterator  inIterator;
    bool modified = false;


    //
    // 1. EXCLUDE_DATE: 'occurrences to delete'.
    //
    if (excludeDates.size() > 0) {
        exIterator = excludeDates.begin();

        while (exIterator != excludeDates.end()) {
            excludeDate = *exIterator;
            if (!excludeDate.size()) {
                continue;
            }

            DATE date = NULL;
            systemTimeDateToDouble(excludeDate, &date);

            cOcc = new ClientAppException();
            cOcc->setOriginalDate(date);
            cOcc->setDeleted(TRUE);

            // Add the exception to the list of exceptions (will be saved during 'save()').
            cRec->addException(cOcc);

            if (cOcc) {
                delete cOcc;  cOcc = NULL;
            }
            exIterator ++;
        }
    }


    //
    // 2. INCLUDE_DATE: 'occurrences to add' -> create new appointment (unlinked)
    //    (added occurrences are not supperted by Outlook)
    //
    if (includeDates.size() > 0) {
        inIterator = includeDates.begin();

        while (inIterator != includeDates.end()) {
            includeDate = *inIterator;
            if (!includeDate.size()) {
                continue;
            }

            // Mark the recurring item as modified: an 'IncludeDate' is converted here to a new item unlinked
            // and so the item will be sent back to server next time.
            modified = true;

            //
            // Copy all properties from original item.
            //
            ClientItem* cItemNew = cItem->copyItem();

            // Set properties modified.
            ((ClientAppointment*)cItemNew)->clearRecPattern();                              // Not recurring, please...
            cItemNew->setProperty(L"IsRecurring", L"0");
            cItemNew->setProperty(L"Start", includeDate);                                   // Set the (different) Start property.
            cItemNew->setProperty(L"AllDayEvent", cItem->getProperty(L"AllDayEvent"));      // Seems necessary to set this property again... (bug?)
            // *** TODO: shift and set the "End" field? ***

            //
            // Save the new appointment to Outlook.
            //
            if (cItemNew->saveItem()) {
                sprintf(lastErrorMsg, ERR_ITEM_SAVE, APPOINTMENT, getSafeItemName(cItemNew).c_str(), cItemNew->getParentPath().c_str());
                goto error;
            }

            // Note: this new appointment will be automatically sent to server next
            //       sync as a NEW item.
            LOG.debug(DBG_ITEM_CREATED_FROM_EXCEPTION, getSafeItemName(cItemNew).c_str());

            if (cItemNew) {
                delete cItemNew;
                cItemNew = NULL;
            }
            inIterator ++;
        }
    }


    //
    // 3. If this appointment has been modified, we MUST add it to the list 
    //    of MOD items on next sync! (use a list of forced modified items)
    //
    if (modified) {

        // Open the file to store IDs for next sync
        WCHAR* filePath = readDataPath(APPOINTMENT_FORCED_MODIFIED);
        if (!filePath) {
            sprintf(lastErrorMsg, ERR_WFILE_OPEN, APPOINTMENT_FORCED_MODIFIED);
            goto error;
        }

        // MUST first save the item - to have the item ID...
        if (cItem->saveItem()) {
            sprintf(lastErrorMsg, ERR_ITEM_SAVE, L"recurring appointment", getSafeItemName(cItem).c_str(), cItem->getParentPath().c_str());
            goto error;
        }

        // This item is added to a list of additional MOD items for next sync.
        wstring xml = L"<Item>";
        xml += cItem->getID();
        xml += L"</Item>\n";
        if (writeToFile(xml, filePath, L"a")) {     // Append at the end of file.
            goto error;
        }

        if (filePath) {
            delete [] filePath;
            filePath = NULL;
        }
    }

    return 0;

error:
    LOG.error(lastErrorMsg);
    return -1;
}


