/*
* Copyright (C) 2003-2007 Funambol
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "base/fscapi.h"
#include "base/Log.h"
#include "winmaincpp.h"
#include "outlook/defs.h"
#include "utils.h"

#include "outlook/ClientAppointment.h"
#include "outlook/ClientException.h"
#include "outlook/utils.h"
#include "outlook/itemProps.h"

using namespace std;



/* 
 * Constructor:
 * Initialize class members.
 * Instance the Redemption COM pointer.
*/
ClientAppointment::ClientAppointment() : ClientItem() {

    pAppointment     = NULL;
    pSafeAppointment = NULL;

    createSafeAppointmentInstance();
}

// Destructor
ClientAppointment::~ClientAppointment() {
}



/*
 * Set a COM pointer to this object.
 * ------------------------------------
 * This method is used to link the object to the correspondent
 * outlook COM pointer. All class members are overwrited by this call.
 * The method MUST be called before using this object, as the constructor
 * doesn't link the class COM pointer.
 * This method also links Redemption COM pointer to the item.
 * If 'itemID' parameter is an empty string, then ID is derived
 * from the entryID of the outlook item.
 */
void ClientAppointment::setCOMPtr(_AppointmentItemPtr& ptr, const wstring& itemID) {

    pAppointment = ptr;

    try {
        pSafeAppointment->Item = pAppointment;

        pItemProperties = pAppointment->ItemProperties;
        propertiesCount = pItemProperties->Count;
        if (propertiesCount) {
            pItemProperty = pItemProperties->Item(0);
        }

        // ID passed as parameter
        if (itemID != EMPTY_WSTRING) {
            ID = itemID;
        }
        // ID retrieved from Outlook
        else {
            _bstr_t bstrID = pAppointment->GetEntryID();
            // ID could not exist (if item not saved)
            if (bstrID.GetBSTR()) {
                ID = (WCHAR*)bstrID;
            }
            else {
                ID = EMPTY_WSTRING;
            }
        }
        // Rec Pattern
        initializeRecPattern();

        // full path of parent folder
        MAPIFolderPtr parentFolder = (MAPIFolderPtr)pAppointment->GetParent();
        if (parentFolder) {
            parentPath = (WCHAR*)parentFolder->GetFullFolderPath();
        }
    }

    catch(_com_error &e) {
        manageComErrors(e);
        throwClientFatalException(ERR_OUTLOOK_ITEM_ASSIGN);
    }

    hr              = S_OK;
    itemType        = APPOINTMENT;
    propertiesIndex = 0;

    // Assume all items of this type have same props!
    //propertyMap.clear();
}


// Here ID is derived from the entryID of the item.
void ClientAppointment::setCOMPtr(_AppointmentItemPtr& ptr) {
    return setCOMPtr(ptr, EMPTY_WSTRING);
}



/*
 * Creates an instance for the Redemption COM pointer.
 */
void ClientAppointment::createSafeAppointmentInstance() {

    try {
        hr = pSafeAppointment.CreateInstance(RED_SAFEAPPOINTMENT);
        if (FAILED(hr)) {
            throwClientFatalException(ERR_OUTLOOK_SAFEITEM);
        }
    }
    catch(_com_error &e) {
        manageComErrors(e);
        throwClientFatalException(ERR_OUTLOOK_SAFEITEM);
    }
}


/*
 * Initialize Recurrence Pattern object.
 *   - initialize members
 *   - link COM Ptr (if exists)
 *   - set infos for the change-day of rec pattern
 */
void ClientAppointment::initializeRecPattern() {

    RecurrencePatternPtr rp;

    //
    // This is an item retrieved (not empty).
    //
    if (ID != EMPTY_WSTRING) {
        BOOL isRec = vBoolToBOOL(pAppointment->GetIsRecurring());
        
        //rp = pAppointment->GetRecurrencePattern();
        //if (isRec)  recPattern.setRecurrence  ();
        //else        recPattern.clearRecurrence();

        if (isRec) {
            rp = pAppointment->GetRecurrencePattern();
            recPattern.setRecurrence();
        }
        else {
            rp = NULL;
            recPattern.clearRecurrence();
        }

        // This is used for the change-day option
        recPattern.setStart(getComplexProperty(L"Start"));
    }

    //
    // This is a new empty item (add): rec pattern not exist yet!!
    // Note: calling 'getRecPattern()' will lead to errors...
    //
    else {
        rp = NULL;
        recPattern.clearRecurrence();
    }

    //
    // Link the COM Ptr (NULL if new item)
    //
    recPattern.setCOMPtr(rp);
}



/**
 * Returns a pointer to the internal ClientRecurrence object 
 * Returns NULL if it's not recurring.
 */
ClientRecurrence* ClientAppointment::getRecPattern() {

    bool isRec = recPattern.isRecurring();
    return (isRec? &recPattern : NULL);
}



/**
 * Used to clear recurrence pattern. 
 * Also removes any appointment exception (if exists).
 * Throws a ClientException if operation fails.
 * Returns 0 if no errors.
 */
int ClientAppointment::clearRecPattern() {

    try {
        hr = pAppointment->ClearRecurrencePattern();
        if (FAILED(hr)) {
            goto error;
        }
        recPattern.clearRecurrence();
    }
    catch(_com_error &e) {
        manageComErrors(e);
        goto error;
    }
    return 0;

error: 
    sprintf(lastErrorMsg, ERR_OUTLOOK_CLEAR_REC, itemType.c_str(), getSafeItemName(this).c_str());
    throwClientException(lastErrorMsg);
    return 1;
}




/*
 * Save the current appointment item.
 *
 * Notes:
 * - exceptions can be saved only if the appointment is already saved
 * - saving the recurrence pattern results in losing all exceptions (if any),
 *   so we must save the appointment before saving exceptions
 *
 * @return: 0 if no errors.
 */
int ClientAppointment::saveItem() {
    
    try {
        // 1. Save recurrence pattern if needed.
        if (recPattern.isRecurring()) {
            if (recPattern.save()) {
                goto error;
            }
        }

        // 2. Save the appointment item.
        hr = pAppointment->Save();
        if (FAILED(hr)) {
            goto error;
        }

        // 3. Save exceptions (MUST: after saving the item).
        if (recPattern.saveAllExceptions()) {
            goto error;
        }

        // Get the new entry-ID.
        ID = (WCHAR*)pAppointment->GetEntryID();
    }
    catch(_com_error &e) {
        manageComErrors(e);
        goto error;
    }

    return 0;

error:
    sprintf(lastErrorMsg, ERR_OUTLOOK_ITEM_SAVE, itemType.c_str());
    throwClientException(lastErrorMsg);
    return 1;
}


/*
 * Delete the current Item.
 * @return: 0 if no errors.
 */
int ClientAppointment::deleteItem() {
    
    try {
        hr = pAppointment->Delete();
        if (FAILED(hr)) {
            goto error;
        }
    }
    catch(_com_error &e) {
        manageComErrors(e);
        goto error;
    }

    return 0;

error:
    sprintf(lastErrorMsg, ERR_OUTLOOK_ITEM_DELETE, itemType.c_str(), ID.c_str());
    throwClientException(lastErrorMsg);
    return 1;
}



/**
 * Returns a (new allocated) copy of this item.
 * @return : a pointer to the new ClientItem object (must bee deleted by the caller).
 *           NULL in case of errors.
 */
ClientItem* ClientAppointment::copyItem() {

    IDispatchPtr        pNew    = NULL;
    _AppointmentItemPtr pAppNew = NULL;
    ClientAppointment*  cNew    = NULL;

    if (!pAppointment) {
        goto error;
    }

    try {
        // Copy the COM pointer
        pNew = pAppointment->Copy();
        if (!pNew) {
            goto error;
        }
        pAppNew = (_AppointmentItemPtr)pNew;
        if (!pAppNew) {
            goto error;
        }
    }
    catch(_com_error &e) {
        manageComErrors(e);
        goto error;
    }

    // Set the new COM pointer to a new allocated ClientAppointment.
    cNew = new ClientAppointment();
    cNew->setCOMPtr(pAppNew);

    return (ClientItem*)cNew;

error:
    return NULL;
}





// Always retrieve the updated property.
BOOL ClientAppointment::isAllDayEvent() {
    VARIANT_BOOL allday = pAppointment->GetAllDayEvent();
    return vBoolToBOOL(allday);
}


//
// ------------------------------ Methods to manage item properties -----------------------------
//

/*
 * Return true if the property is protected by Outlook Security patch.
 * Protected properties are listed in safeCalendarProps array.
 */
bool ClientAppointment::isSecureProperty(const wstring& propertyName) {
    
    for (int i=0; safeCalendarProps[i]; i++) {
        if (propertyName == safeCalendarProps[i]) {
            return true;
        }
    }
    return false;
}

/*
 * Return true if the property needs some specific conversion of data.
 * Complex properties are listed in complexCalendarProps array.
 * Recurrence properties are listed in recurrenceProps array.
 */
bool ClientAppointment::isComplexProperty(const wstring& propertyName) {
    
    for (int i=0; complexCalendarProps[i]; i++) {
        if (propertyName == complexCalendarProps[i]) {
            return true;
        }
    }
    for (int i=0; recurrenceProps[i]; i++) {
        if (propertyName == recurrenceProps[i]) {
            return true;
        }
    }
    return false;
}



// ------------------------------- GET PROPERTY -------------------------------
/*
 * Return the Item property value from the property name.
 * Use appropriate Redemption function for each property.
 * Note: this method uses the redemption library.
 *
 * @param propertyName  : the name of the property
 * @return              : the value retrieved (from Outlook)
 */
const wstring ClientAppointment::getSafeProperty(const wstring& propertyName) {
    
    wstring propertyValue = EMPTY_WSTRING;
    BSTR tmpVal;

    if (propertyName == L"Body") {
        hr = pSafeAppointment->get_Body(&tmpVal);
    }
    else if (propertyName == L"OptionalAttendees") {
        hr = pSafeAppointment->get_OptionalAttendees(&tmpVal);
    }

    if (FAILED(hr)) {
        sprintf(lastErrorMsg, ERR_OUTLOOK_PROP_VALUE, itemType.c_str(), propertyName.c_str());
        throwClientException(lastErrorMsg);
    }
    if (tmpVal) {
        propertyValue = tmpVal;
    }

    return propertyValue;
}




/*
 * Return the Item property value from the property name.
 * Here are processed properties that need specific conversion of 
 * the property value. Also recurrence properties are retrieved here.
 *
 * @param propertyName  : the name of the property
 * @return              : the value retrieved (from Outlook)
 */
const wstring ClientAppointment::getComplexProperty(const wstring& propertyName) {

    wstring propertyValue = EMPTY_WSTRING;
    WCHAR tmp[64];
    VARIANT_BOOL vbool;
    DATE date;

    //
    // Date
    //
    if (propertyName == L"Start") {                             // All day: "yyyy-MM-dd"
        date = pAppointment->GetStart();                        // else   : "YYYYMMDDThhmmssZ"
        if (isAllDayEvent()) {
            doubleToSystemTimeDate(propertyValue, date);
        }
        else {
            doubleToSystemTime(propertyValue, date, USE_UTC);
        }
    }
    else if (propertyName == L"End") {                          // All day: "yyyy-MM-dd"
        date = pAppointment->GetEnd();                          // else   : "YYYYMMDDThhmmssZ"
        if (isAllDayEvent()) {
            // Correction for All-day event (-1 day)
            date -= 1;
            doubleToSystemTimeDate(propertyValue, date);
        }
        else {
            doubleToSystemTime(propertyValue, date, USE_UTC);
        }
    }
    else if (propertyName == L"ReplyTime") {                    // "YYYYMMDDThhmmssZ"
        date = pAppointment->GetReplyTime();
        if (date && date < LIMIT_MAX_DATE) {
            doubleToSystemTime(propertyValue, date, USE_UTC);
        }
        else {
            propertyValue = EMPTY_WSTRING;
        }
    }
    else if (propertyName == L"LastModificationTime") {         // "(double format)"
        date = pAppointment->GetLastModificationTime();
        swprintf_s(tmp, L"%.12f", date);
        propertyValue = tmp;
    }

    //
    // Boolean
    //
    else if (propertyName == L"AllDayEvent") {
        wsprintf(tmp, TEXT("%d"), isAllDayEvent());
        propertyValue = tmp;
    }
    else if (propertyName == L"IsRecurring") {
        vbool = pAppointment->GetIsRecurring();
        wsprintf(tmp, TEXT("%d"), vBoolToBOOL(vbool));
        propertyValue = tmp;    
    }
    else if (propertyName == L"NoAging") {
        vbool = pAppointment->GetNoAging();
        wsprintf(tmp, TEXT("%d"), vBoolToBOOL(vbool));
        propertyValue = tmp;   
    }
    else if (propertyName == L"ReminderSet") {
        vbool = pAppointment->GetReminderSet();
        wsprintf(tmp, TEXT("%d"), vBoolToBOOL(vbool));
        propertyValue = tmp;   
    }
    else if (propertyName == L"UnRead") {
        vbool = pAppointment->GetUnRead();
        wsprintf(tmp, TEXT("%d"), vBoolToBOOL(vbool));
        propertyValue = tmp;
    }

    //
    // Recurrence Pattern: get from internal object (recPattern).
    //
    else {
        if (recPattern.isRecurring()) {
            propertyValue = getRecurrenceProperty(propertyName); 
        }
        else {
            return EMPTY_WSTRING;
        }
    }

    return propertyValue;
}


/*
 * Return the recurrence pattern property value from the property name.
 *
 * Properties are retrieved from internal object (recPattern) as a buffer.
 * Rec pattern values must be first retrieved all together, this is done
 * internally in ClientRecurrence when the first property is retrieved.
 * If UTC is used, rec props are converted inside ClientRecurrence.
 *
 * @param propertyName  : the name  of the property
 * @return              : the value retrieved from recPattern (buffer)
 */
const wstring ClientAppointment::getRecurrenceProperty(const wstring& propertyName) {
    
    wstring propertyValue = EMPTY_WSTRING;
    WCHAR tmp[64];

    if (propertyName == L"RecurrenceType") {
        wsprintf(tmp, TEXT("%d"), recPattern.getRecurrenceType());
        propertyValue = tmp;
    }
    else if (propertyName == L"Interval") {
        wsprintf(tmp, TEXT("%d"), recPattern.getInterval());
        propertyValue = tmp;
    }
    else if (propertyName == L"MonthOfYear") {                              // Set EMPTY if = 0 ?
        wsprintf(tmp, TEXT("%d"), recPattern.getMonthOfYear());
        propertyValue = tmp;
    }
    else if (propertyName == L"DayOfMonth") {                               // Set EMPTY if = 0 ?
        wsprintf(tmp, TEXT("%d"), recPattern.getDayOfMonth());
        propertyValue = tmp;
    }
    else if (propertyName == L"DayOfWeekMask") {                            // Set EMPTY if = 0 ?
        wsprintf(tmp, TEXT("%d"), recPattern.getDayOfWeekMask());
        propertyValue = tmp;
    }
    else if (propertyName == L"Instance") {
        wsprintf(tmp, TEXT("%d"), recPattern.getInstance());
        propertyValue = tmp;
    }
    else if (propertyName == L"PatternStartDate") {
        propertyValue = recPattern.getPatternStartDate();
    }
    else if (propertyName == L"NoEndDate") {
        wsprintf(tmp, TEXT("%d"), recPattern.getNoEndDate());
        propertyValue = tmp;
    }
    else if (propertyName == L"PatternEndDate") {
        propertyValue = recPattern.getPatternEndDate();
    }
    else if (propertyName == L"Occurrences") {
        int occ = recPattern.getOccurrences();
        if (occ > 0) {
            wsprintf(tmp, TEXT("%d"), occ);
            propertyValue = tmp;
        }
        else {
            // for compatibility
            propertyValue = EMPTY_WSTRING;
        }
    }
    else {
        sprintf(lastErrorMsg, ERR_OUTLOOK_PROP_NOT_FOUND, propertyName.c_str(), itemType.c_str());
        throwClientException(lastErrorMsg);
        return EMPTY_WSTRING;
    }

    return propertyValue;
}



// ------------------------------- SET PROPERTY -------------------------------
/*
 * Set the Item property value for the specific property name.
 * Here are processed properties that need specific conversion of the property value. 
 * Also recurrence properties are set from here.
 *
 * @param propertyName  : the name of the property
 * @param propertyValue : the value to store
 * @return              : 0 if no errors, 1 if errors
 */
int ClientAppointment::setComplexProperty(const wstring& propertyName, const wstring& propertyValue) {
    
    int  ret = 0;
    VARIANT_BOOL vbool;

    // Default values used if propertyValue = empty.
    DATE date     = 0;
    int  intValue = 0;

    int len = propertyValue.length();

    //
    // Date
    //
    if (propertyName == L"Start") {
        if (!len) {
            // this is required!
            sprintf(lastErrorMsg, ERR_OUTLOOK_PROP_VALUE_REQUIRED, L"Start", itemType.c_str());
            return 1;
        }
        if (isSystemTimeDate(propertyValue)) {
            systemTimeDateToDouble(propertyValue, &date);
        }
        else {
            systemTimeToDouble(propertyValue, &date);
        }
        hr = pAppointment->put_Start(date);

        // This is used for the change-day option
        recPattern.setStart(propertyValue);
    }

    else if (propertyName == L"End") {
        if (!len) {
            // this is required!
            sprintf(lastErrorMsg, ERR_OUTLOOK_PROP_VALUE_REQUIRED, L"End", itemType.c_str());
            return 1;
        } 
        if (isSystemTimeDate(propertyValue)) {
            systemTimeDateToDouble(propertyValue, &date);
            // Correction for All-day event (+1 day)
            date += 1;
        }
        else {
            systemTimeToDouble(propertyValue, &date);
        }
        hr = pAppointment->put_End(date);
    }

    else if (propertyName == L"ReplyTime") {
        date = 0;
        if (len) {
            systemTimeDateToDouble(propertyValue, &date);             // *** verify time 00:00 is correct? ***
        }
        hr = pAppointment->put_ReplyTime(date);                       // Use '0' to clear
    }

    //
    // Boolean
    //
    else if (propertyName == L"AllDayEvent") {
        intValue = 1;   // as a default
        if (len) {
            intValue = _wtoi(propertyValue.c_str());
        }
        vbool = BOOLToVBool(intValue);
        hr = pAppointment->put_AllDayEvent(vbool);
    }

    else if (propertyName == L"IsRecurring") {
        BOOL isRec = FALSE;   // as a default
        if (len) {
            isRec = _wtoi(propertyValue.c_str());
        }
        // MUST set the 'recurring' value and link COM Ptr.
        if (isRec) {
            recPattern.setRecurrence();
            recPattern.setCOMPtr(pAppointment->GetRecurrencePattern());
        }
        else {
            recPattern.clearRecurrence();
        }
    }

    else if (propertyName == L"NoAging") {
        if (len) {
            intValue = _wtoi(propertyValue.c_str());
        }
        vbool = BOOLToVBool(intValue);
        hr = pAppointment->put_NoAging(vbool);
    }

    else if (propertyName == L"ReminderSet") {
        if (len) {
            intValue = _wtoi(propertyValue.c_str());
        }
        vbool = BOOLToVBool(intValue);
        hr = pAppointment->put_ReminderSet(vbool); 
    }

    else if (propertyName == L"UnRead") {
        if (len) {
            intValue = _wtoi(propertyValue.c_str());
        }
        vbool = BOOLToVBool(intValue);
        hr = pAppointment->put_UnRead(vbool); 
    }

    //
    // Recurrence Pattern: set internal object values (recPattern).
    //
    else {
        // Only non-empty values
        if (len) {
            ret = setRecurrenceProperty(propertyName, propertyValue);
        }
    }

    if (FAILED(hr)) {
        sprintf(lastErrorMsg, ERR_OUTLOOK_PROP_VALUE_SET, propertyName.c_str(), propertyValue.c_str(), itemType.c_str());
        throwClientException(lastErrorMsg);
        ret = 1;
    }
    return ret;
}



/*
 * Set Recurrence Pattern property: set internal object values (recPattern).
 * All rec pattern must be saved together at the end calling 'save()'
 * method of ClientRecurrence. This is because:
 *   - rec props MUST be verified consistent each others
 *   - if UTC is used, rec props are first converted inside recPattern.
 *
 * @param propertyName  : the name of the property
 * @param propertyValue : the value to store
 * @return              : 0 if no errors, 1 if errors
 */
int ClientAppointment::setRecurrenceProperty(const wstring& propertyName, const wstring& propertyValue) {
    int intValue = 0;

    if (propertyName == L"RecurrenceType") {
        intValue = _wtoi(propertyValue.c_str());
        recPattern.setRecurrenceType(intValue);
    }
    else if (propertyName == L"Interval") {
        intValue = _wtoi(propertyValue.c_str());
        recPattern.setInterval(intValue);
    }
    else if (propertyName == L"MonthOfYear") {
        intValue = _wtoi(propertyValue.c_str());
        recPattern.setMonthOfYear(intValue);
    }
    else if (propertyName == L"DayOfMonth") {
        intValue = _wtoi(propertyValue.c_str());
        recPattern.setDayOfMonth(intValue);
    }
    else if (propertyName == L"DayOfWeekMask") {
        intValue = _wtoi(propertyValue.c_str());
        recPattern.setDayOfWeekMask(intValue);
    }
    else if (propertyName == L"Instance") {
        intValue = _wtoi(propertyValue.c_str());
        recPattern.setInstance(intValue);
    }
    else if (propertyName == L"PatternStartDate") {
        recPattern.setPatternStartDate(propertyValue);
    }
    else if (propertyName == L"NoEndDate") {
        BOOL bValue = _wtoi(propertyValue.c_str());
        recPattern.setNoEndDate(bValue);
    }
    else if (propertyName == L"PatternEndDate") {
        recPattern.setPatternEndDate(propertyValue);
    }
    else if (propertyName == L"Occurrences") {
        intValue = _wtoi(propertyValue.c_str());
        recPattern.setOccurrences(intValue);
    }
    else {
        sprintf(lastErrorMsg, ERR_OUTLOOK_PROP_NOT_FOUND, propertyName.c_str(), itemType.c_str());
        throwClientException(lastErrorMsg);
        return 1;
    }

    return 0;
}





/* 
 * Copy Constructor
 */
ClientAppointment::ClientAppointment(const ClientAppointment& c) {

    hr              = c.hr;
    ID              = c.ID;
    itemType        = c.itemType;
    parentPath      = c.parentPath;
    pAppointment    = c.pAppointment;
    pItemProperties = c.pItemProperties;
    pItemProperty   = c.pItemProperty;
    propertiesIndex = c.propertiesIndex;
    propertiesCount = c.propertiesCount;
    propertyMap     = c.propertyMap;

    createSafeAppointmentInstance();
    pSafeAppointment= c.pSafeAppointment;
    recPattern      = c.recPattern;
}


/* 
 * Operator =
*/
ClientAppointment ClientAppointment::operator=(const ClientAppointment& c) {

    ClientAppointment cnew;

    cnew.hr              = c.hr;
    cnew.ID              = c.ID;
    cnew.itemType        = c.itemType;
    cnew.parentPath      = c.parentPath;
    cnew.pAppointment    = c.pAppointment;
    cnew.pItemProperties = c.pItemProperties;
    cnew.pItemProperty   = c.pItemProperty;
    cnew.propertiesIndex = c.propertiesIndex;
    cnew.propertiesCount = c.propertiesCount;
    cnew.propertyMap     = c.propertyMap;

    cnew.createSafeAppointmentInstance();
    cnew.pSafeAppointment= c.pSafeAppointment;
    cnew.recPattern      = c.recPattern;

    return cnew;
}


//void ClientAppointment::test() {
//
//    HRESULT hres;
//
//
//    try {
//        
//        hres = pAppointment->put_AllDayEvent(VARIANT_TRUE);
//        hres = pAppointment->put_Start(39010.000000);
//        hres = pAppointment->put_End  (39011.000000);
//        //hres = pAppointment->ClearRecurrencePattern();
//        pAppointment->PutSubject(L"Test");
//        pAppointment->PutBusyStatus(olBusy);
//        pAppointment->PutImportance(olImportanceLow);
//        pAppointment->PutMeetingStatus(olNonMeeting);
//        pAppointment->PutReminderMinutesBeforeStart(0);
//        pAppointment->PutReminderSet(VARIANT_FALSE);
//        pAppointment->PutSensitivity(olNormal);
//
//
//        //pAppointment->Display();
//
//        hres = pAppointment->Save();
//
//
//        DATE st, en;
//        VARIANT_BOOL ad;
//        _bstr_t sj;
//        hres = pAppointment->get_Start(&st);
//        hres = pAppointment->get_End  (&en);
//        RecurrencePatternPtr p = pAppointment->GetRecurrencePattern();
//        hres = pAppointment->get_AllDayEvent(&ad);
//        sj = pAppointment->GetSubject();
//        LOG.debug("test done");
//    }
//
//    catch(_com_error &e) {
//        manageComErrors(e);
//    }
//}


