/*
* Copyright (C) 2003-2007 Funambol
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "base/fscapi.h"
#include "base/Log.h"
#include "winmaincpp.h"
#include "outlook/defs.h"

#include "outlook/ClientApplication.h"
#include "outlook/ClientContact.h"
#include "outlook/ClientException.h"
#include "outlook/utils.h"
#include "outlook/itemProps.h"


using namespace std;



/* 
 * Constructor:
 * Dummy initialize class members.
 * Instance the Redemption COM pointer.
*/
ClientContact::ClientContact() : ClientItem() {

    pContact        = NULL;
    pSafeContact    = NULL;
    willCreateAnniversaryEvent = false;
    willCreateBirthdayEvent    = false;

    createSafeContactInstance();
}


// Destructor
ClientContact::~ClientContact() {
}


/*
 * Set a COM pointer to this object.
 * ------------------------------------
 * This method is used to link the object to the correspondent
 * outlook COM pointer. All class members are overwrited by this call.
 * The method MUST be called before using this object, as the constructor
 * doesn't link the class COM pointer.
 * This method also links Redemption COM pointer to the item.
 * If 'itemID' parameter is an empty string, then ID is derived
 * from the entryID of the outlook item.
 */
void ClientContact::setCOMPtr(_ContactItemPtr& ptr, const wstring& itemID) {

    pContact = ptr;

    try {
        pSafeContact->Item = pContact;

        pItemProperties = pContact->ItemProperties;
        propertiesCount = pItemProperties->Count;
        if (propertiesCount) {
            pItemProperty = pItemProperties->Item(0);
        }
        // ID passed as parameter
        if (itemID != EMPTY_WSTRING) {
            ID = itemID;
        }
        // ID retrieved from Outlook
        else {
            _bstr_t bstrID = pContact->GetEntryID();
            // ID could not exist (if item not saved)
            if (bstrID.GetBSTR()) {
                ID = (WCHAR*)bstrID;
            }
            else {
                ID = EMPTY_WSTRING;
            }
        }
        // full path of parent folder
        MAPIFolderPtr parentFolder = (MAPIFolderPtr)pContact->GetParent();
        parentPath = (WCHAR*)parentFolder->GetFullFolderPath();
    }
    catch(_com_error &e) {
        manageComErrors(e);
        throwClientFatalException(ERR_OUTLOOK_ITEM_ASSIGN);
    }

    hr              = S_OK;
    itemType        = CONTACT;
    propertiesIndex = 0;
    willCreateAnniversaryEvent = false;
    willCreateBirthdayEvent    = false;

    // Assume all items of this type have same props!
    //propertyMap.clear();
}


// Entry-ID is derived from the entryID of the item.
void ClientContact::setCOMPtr(_ContactItemPtr& ptr) {
    return setCOMPtr(ptr, EMPTY_WSTRING);
}



/*
 * Creates an instance for the Redemption COM pointer.
 */
void ClientContact::createSafeContactInstance() {

    try {
        hr = pSafeContact.CreateInstance(RED_SAFECONTACT);
        if (FAILED(hr)) {
            throwClientFatalException(ERR_OUTLOOK_SAFEITEM);
        }
    }
    catch(_com_error &e) {
        manageComErrors(e);
        throwClientFatalException(ERR_OUTLOOK_SAFEITEM);
    }
}




/*
 * Save the current Item.
 * @return: 0 if no errors.
 */
int ClientContact::saveItem() {  
    
    try {
        hr = pContact->Save();
        if (FAILED(hr)) {
            goto error;
        }
        // Get the new entry-ID.
        ID = (WCHAR*)pContact->GetEntryID();
    }
    catch(_com_error &e) {
        manageComErrors(e);
        goto error;
    }  

    return 0;

error:
    sprintf(lastErrorMsg, ERR_OUTLOOK_ITEM_SAVE, itemType.c_str());
    throwClientException(lastErrorMsg);
    return 1;
}


/*
 * Delete the current Item.
 * @return: 0 if no errors.
 */
int ClientContact::deleteItem() {
    
    try {
        hr = pContact->Delete();
        if (FAILED(hr)) {
            goto error;
        }
    }
    catch(_com_error &e) {
        manageComErrors(e);
        goto error;
    }

    return 0;

error:
    sprintf(lastErrorMsg, ERR_OUTLOOK_ITEM_DELETE, itemType.c_str(), ID.c_str());
    throwClientException(lastErrorMsg);
    return 1;
}



/**
 * Returns a (new allocated) copy of this item.
 * @return : a pointer to the new ClientItem object (must bee deleted by the caller).
 *           NULL in case of errors.
 */
ClientItem* ClientContact::copyItem() {

    IDispatchPtr     pNew    = NULL;
    _ContactItemPtr  pConNew = NULL;
    ClientContact*   cNew    = NULL;

    if (!pContact) {
        goto error;
    }

    try {
        // Copy the COM pointer
        pNew = pContact->Copy();
        if (!pNew) {
            goto error;
        }
        pConNew = (_ContactItemPtr)pNew;
        if (!pConNew) {
            goto error;
        }
    }
    catch(_com_error &e) {
        manageComErrors(e);
        goto error;
    }

    // Set the new COM pointer to a new allocated ClientContact.
    cNew = new ClientContact();
    cNew->setCOMPtr(pConNew);

    return (ClientItem*)cNew;

error:
    return NULL;
}


//
// ------------------------------ Methods to manage item properties -----------------------------
//

/*
 * Return true if the property is protected by Outlook Security patch.
 * Protected properties are listed in safeContactProps array.
 */
bool ClientContact::isSecureProperty(const wstring& propertyName) {

    // search for property in safeContactProps array
    for (int i=0; safeContactProps[i]; i++) {
        if (propertyName == safeContactProps[i]) {
            return true;
        }
    }
    return false;
}


/*
 * Return true if the property needs some specific conversion of data.
 * Complex properties are listed in complexContactProps array.
 */
bool ClientContact::isComplexProperty(const wstring& propertyName) {
    
    // search for property in complexContactProps array
    for (int i=0; complexContactProps[i]; i++) {
        if (propertyName == complexContactProps[i]) {
            return true;
        }
    }
    return false;
}




// ------------------------------- GET PROPERTY -------------------------------
/*
 * Return the Item property value from the property name.
 * Use appropriate Redemption function for each property.
 * Note: this method uses the redemption library.
 */
const wstring ClientContact::getSafeProperty(const wstring& propertyName) {

    wstring propertyValue = EMPTY_WSTRING;
    BSTR tmpVal;
    
    if (propertyName == L"Body") {
        hr = pSafeContact->get_Body(&tmpVal);
    }

    else if (propertyName == L"Email1Address") {
        hr = pSafeContact->get_Email1Address(&tmpVal);
    }
    else if (propertyName == L"Email2Address") {
        hr = pSafeContact->get_Email2Address(&tmpVal);
    }
    else if (propertyName == L"Email3Address") {
        hr = pSafeContact->get_Email3Address(&tmpVal);
    }

    else if (propertyName == L"Email1AddressType") {
        hr = pSafeContact->get_Email1AddressType(&tmpVal);
    }
    else if (propertyName == L"Email2AddressType") {
        hr = pSafeContact->get_Email2AddressType(&tmpVal);
    }
    else if (propertyName == L"Email3AddressType") {
        hr = pSafeContact->get_Email3AddressType(&tmpVal);
    }
    // Not used by now...
    //else if (propertyName == L"Email1DisplayName") {
    //    hr = pSafeContact->get_Email1DisplayName(&tmpVal);
    //}
    //else if (propertyName == L"Email2DisplayName") {
    //    hr = pSafeContact->get_Email2DisplayName(&tmpVal);
    //}
    //else if (propertyName == L"Email3DisplayName") {
    //    hr = pSafeContact->get_Email3DisplayName(&tmpVal);
    //}
    else if (propertyName == L"IMAddress") {
        hr = pSafeContact->get_IMAddress(&tmpVal);
    }


    if (FAILED(hr)) {
        sprintf(lastErrorMsg, ERR_OUTLOOK_PROP_VALUE, itemType.c_str(), propertyName.c_str());
        throwClientException(lastErrorMsg);
    }

    if (tmpVal) {
        propertyValue = tmpVal;

        //
        // Manage correction of EX->SMTP address: always return SMTP address automatically.
        //
        if ( (propertyName == L"Email1Address" && pSafeContact->GetEmail1AddressType() == (_bstr_t)L"EX") ||
             (propertyName == L"Email2Address" && pSafeContact->GetEmail2AddressType() == (_bstr_t)L"EX") ||
             (propertyName == L"Email3Address" && pSafeContact->GetEmail3AddressType() == (_bstr_t)L"EX") ) {

            ClientApplication* ol = ClientApplication::getInstance();
            propertyValue = ol->getSMTPfromEX(propertyValue);
        }
        else if (propertyName == L"Email1AddressType" ||
                 propertyName == L"Email2AddressType" ||
                 propertyName == L"Email3AddressType" ) {
            if (propertyValue == L"EX") {
                propertyValue = L"SMTP";
            }
        }
    }

    return propertyValue;
}


/*
 * Return the Item property value from the property name.
 * Here are processed properties that need specific conversion of 
 * the property value.
 */
const wstring ClientContact::getComplexProperty(const wstring& propertyName) {
 
    wstring propertyValue = EMPTY_WSTRING;
    WCHAR tmp[32];
    DATE date;

    //
    // Name: correct first char...
    //
    if (propertyName == L"FirstName" ||
        propertyName == L"LastName") {
        WCHAR* fullName = (WCHAR*)pContact->GetFullName();
        wstring name    = getSimpleProperty(propertyName);
        if (fullName) {
            propertyValue = getCorrectNameValue(name, fullName);
        }
        else {
            propertyValue = name;
        }
    }
    
    //
    // Date
    //
    if (propertyName == L"Anniversary") {
        date = pContact->GetAnniversary();
        if (date < LIMIT_MAX_DATE) {
            doubleToSystemTimeDate(propertyValue, date);                // "yyyy-MM-dd"
        }
    }

    else if (propertyName == L"Birthday") {
        date = pContact->GetBirthday();
        if (date < LIMIT_MAX_DATE) {
            doubleToSystemTimeDate(propertyValue, date);                // "yyyy-MM-dd"
        }
    }

    else if (propertyName == L"LastModificationTime") {                 // "(double format)"
        date = pContact->GetLastModificationTime();
        swprintf_s(tmp, L"%.12f", date);
        propertyValue = tmp;
    }

    //
    // Boolean
    //
    // Not used by now...
    //else if (propertyName == L"NoAging") {
    //    vbool = pContact->GetNoAging();
    //    wsprintf(tmp, TEXT("%d"), vBoolToBOOL(vbool));
    //    propertyValue = tmp;
    //}
    // Not used by now...
    //else if (propertyName == L"UnRead") {
    //    vbool = pContact->GetUnRead();
    //    wsprintf(tmp, TEXT("%d"), vBoolToBOOL(vbool));
    //    propertyValue = tmp;
    //}

    return propertyValue;
}



// ------------------------------- SET PROPERTY -------------------------------
/*
 * Set the Item property value for the specific property name.
 * Here are processed properties that need specific conversion of the property value. 
 *
 * @param propertyName  : the name of the property
 * @param propertyValue : the value to store
 * @return              : 0 if no errors, 1 if errors
 */
int ClientContact::setComplexProperty(const wstring& propertyName, const wstring& propertyValue) {

    DATE date;
 
    //
    // Name: manage first char...
    //
    if (propertyName == L"FirstName" ||
        propertyName == L"LastName") {
        // Not an error!!! Set 2 times to avoid automatic 
        // replace of first char to upper case! (funny, uh?)
        setSimpleProperty(propertyName, propertyValue);
        setSimpleProperty(propertyName, propertyValue);
    }

    //
    // Date
    //
    if (propertyName == L"Anniversary") {                   // Expected "yyyy-MM-dd" (UTC also accepted)
        systemTimeDateToDouble(propertyValue, &date);       // Use REFERRED_MAX_DATE to clear
        hr = pContact->put_Anniversary(date);
        
        if (SUCCEEDED(hr) && date < LIMIT_MAX_DATE) {       // Outlook will create automatically an event!
            willCreateAnniversaryEvent = true;
        }
    }

    else if (propertyName == L"Birthday") {                 // Expected "yyyy-MM-dd" (UTC also accepted)
        systemTimeDateToDouble(propertyValue, &date);       // Use REFERRED_MAX_DATE to clear
        hr = pContact->put_Birthday(date);

        if (SUCCEEDED(hr) && date < LIMIT_MAX_DATE) {       // Outlook will create automatically an event!
            willCreateBirthdayEvent = true;
        }
    }

    //
    // Boolean
    //
    // Not used by now...
    //else if (propertyName == L"NoAging") {
    //    intValue = 0;    // as a default
    //    if (propertyValue.length() > 0) {
    //        intValue = _wtoi(propertyValue.c_str());
    //    }
    //    vbool = BOOLToVBool(intValue);
    //    hr = pContact->put_NoAging(vbool);
    //}
    // Not used by now...
    //else if (propertyName == L"UnRead") {
    //    intValue = 0;    // as a default
    //    if (propertyValue.length() > 0) {
    //        intValue = _wtoi(propertyValue.c_str());
    //    }
    //    vbool = BOOLToVBool(intValue);
    //    hr = pContact->put_UnRead(vbool); 
    //}


    if (FAILED(hr)) {
        sprintf(lastErrorMsg, ERR_OUTLOOK_PROP_VALUE_SET, propertyName.c_str(), propertyValue.c_str(), itemType.c_str());
        throwClientException(lastErrorMsg);
        return 1;
    }
    return 0;
}




// ------------------------------- Other methods -------------------------------


// Return true if Anniversary event has been created by Outlook during saveItem().
// Outlook creates this event automatically if Anniversary is set.
bool ClientContact::createdAnniversaryEvent() {

    BOOL saved = vBoolToBOOL(pContact->GetSaved());
    if (saved && willCreateAnniversaryEvent) {
        return true;
    }
    else {
        return false;
    }
}

// Return true if Birthday event has been created by Outlook during saveItem().
// Outlook creates this event automatically if Birthday is set.
bool ClientContact::createdBirthdayEvent() {

    BOOL saved = vBoolToBOOL(pContact->GetSaved());
    if (saved && willCreateBirthdayEvent) {
        return true;
    }
    else {
        return false;
    }
}



/*
 * Internal utility: get the correct format of name fields.
 * Microsoft Outlook automatically converts the first character of some
 * fields into upper-case, even if they are in lower-case.
 * So this function tries to extract these two fields from 'fullName' field, because
 * here the correct format is preserved.
 * NOTE:
 * 'FirstName' and 'LastName' are key values for synchronization and one
 * wrong character could lead to duplicates.
 */
const wstring ClientContact::getCorrectNameValue(const wstring& name, const wstring& fullName) {

    wstring::size_type start;

    //
    // First search the name as passed.
    //
    start = fullName.find(name, 0);
    if (start != wstring::npos) {
        // found!
        return name;
    }
    
    //
    // Get name with 1st letter lower case.
    //
    wstring lowerName = name;
    lowerName[0] = towlower(lowerName[0]);

    // Search the name modified.
    start = fullName.find(lowerName, 0);
    if (start != wstring::npos) {
        // found!
        return lowerName;
    }
    else {
        // not found: default = name.
        return name;
    }
}



/* 
 * Copy Constructor
*/
ClientContact::ClientContact(const ClientContact& c) {
    hr              = c.hr;
    ID              = c.ID;
    itemType        = c.itemType;
    parentPath      = c.parentPath;
    pContact        = c.pContact;
    pItemProperties = c.pItemProperties;
    pItemProperty   = c.pItemProperty;
    propertiesIndex = c.propertiesIndex;
    propertiesCount = c.propertiesCount;
    propertyMap     = c.propertyMap;

    createSafeContactInstance();
    pSafeContact    = c.pSafeContact;

    willCreateAnniversaryEvent = c.willCreateAnniversaryEvent;
    willCreateBirthdayEvent    = c.willCreateBirthdayEvent;
}


/* 
 * Operator =
*/
ClientContact ClientContact::operator=(const ClientContact& c) {

    ClientContact cnew;

    cnew.hr              = c.hr;
    cnew.ID              = c.ID;
    cnew.itemType        = c.itemType;
    cnew.parentPath      = c.parentPath;
    cnew.pContact        = c.pContact;
    cnew.pItemProperties = c.pItemProperties;
    cnew.pItemProperty   = c.pItemProperty;
    cnew.propertiesIndex = c.propertiesIndex;
    cnew.propertiesCount = c.propertiesCount;
    cnew.propertyMap     = c.propertyMap;

    cnew.createSafeContactInstance();
    cnew.pSafeContact    = c.pSafeContact;

    cnew.willCreateAnniversaryEvent = c.willCreateAnniversaryEvent;
    cnew.willCreateBirthdayEvent    = c.willCreateBirthdayEvent;

    return cnew;
}

