/*
* Copyright (C) 2003-2007 Funambol
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "base/fscapi.h"
#include "base/Log.h"
#include "winmaincpp.h"
#include "outlook/defs.h"

#include "outlook/ClientRecurrence.h"
#include "outlook/ClientException.h"
#include "outlook/utils.h"
#include "outlook/itemProps.h"

using namespace std;




// Constructor:
ClientRecurrence::ClientRecurrence() {

    pRec         = NULL;
    start        = EMPTY_WSTRING;

    // not recurring, nor updated.
    clearRecurrence();

    numRecursions = 0;
}


// Destructor
ClientRecurrence::~ClientRecurrence() {

    appExceptions.clear();
}



/*
 * Set a COM pointer to this object.
 * ------------------------------------
 * This method is used to link the object to the correspondent
 * outlook COM pointer.
 * The method MUST be called before using this object, as the constructor
 * doesn't link the class COM pointer.
 *
 * Note: RecPatternPtr could be passed NULL (if new item), so check it
 *       before using it (read() and save()).
 */
void ClientRecurrence::setCOMPtr(RecurrencePatternPtr& ptr) {

    pRec = ptr;

    appExceptions.clear();
    numRecursions = 0;

    // We just linked a different rec pattern, 
    // current members are not updated!
    isUpdated = false;
}





//Return true if the rec pattern is active.
bool ClientRecurrence::isRecurring() {
    return recurring;
}


// Set the recurrence (recurring = true).
void ClientRecurrence::setRecurrence() {
    recurring = true;
}


// Clear the recurrence (recurring = false).
void ClientRecurrence::clearRecurrence() {
    
    recurring = false;
    isUpdated = false;

    // initialize (these values indicate that props are not set)
    recurrenceType   = -1;
    interval         = -1;
    monthOfYear      = -1;
    dayOfMonth       = -1;
    dayOfWeekMask    = -1;
    instance         = -1;
    patternStartDate = EMPTY_WSTRING;
    noEndDate        = FALSE;
    patternEndDate   = EMPTY_WSTRING;
    occurrences      = -1;
}



/////////////////////////////////////////////////////////////////////////
/*
 * Retrieve all properties from Outlook -> set isUpdated = true.
 * If UTC is used, all props are converted to correct values.
 * 
 * @return: 0 if no errors
 */
int ClientRecurrence::read() {

    DATE patternEnd;
    VARIANT_BOOL vNoEnd;


    // Check COM Ptr / Recurrence active
    if (!pRec || !isRecurring()) {
        sprintf(lastErrorMsg, ERR_OUTLOOK_REC_NOT_SET, L"read");
        throwClientException(lastErrorMsg);
        return 1;
    }



    //
    // ---------- Get ALL ------------
    //
    // TBD: add another try/catch (more specific)
    recurrenceType = pRec->GetRecurrenceType();
    interval       = pRec->GetInterval();
    monthOfYear    = pRec->GetMonthOfYear(); 
    dayOfMonth     = pRec->GetDayOfMonth();
    dayOfWeekMask  = pRec->GetDayOfWeekMask();
    instance       = pRec->GetInstance();
    vNoEnd         = pRec->GetNoEndDate();
    patternEnd     = pRec->GetPatternEndDate();
    occurrences    = pRec->GetOccurrences();


    //
    // Get patternStart from start (use exactly the same - new from 6.0.4).
    //
    patternStartDate = EMPTY_WSTRING;
    if (start != EMPTY_WSTRING) {
        patternStartDate = start;
    }
    else {
        // If 'Start' is empty, use the patternStartDate...
        DATE patternStart = pRec->GetPatternStartDate();
        if (patternStart < LIMIT_MAX_DATE) {
            doubleToSystemTimeDate(patternStartDate, patternStart);
        }
    }


    //
    // --------- Conversions ---------
    //
    // yearly/yearNth interval = 1 (for compatibility)
    if (recurrenceType > 4) {
        interval = 1;
    }
    // For compatibility: weekly with interval = 0 is not correct...
    if ((recurrenceType == 1) && (interval == 0)) {
        interval = 1;
    }

    noEndDate = vBoolToBOOL(vNoEnd);

    if (patternEnd < LIMIT_MAX_DATE) {
        doubleToSystemTime(patternEndDate, patternEnd, USE_UTC);            // "YYYYMMDDThhmmssZ" - new from 6.0.4.
    }
    else {
        patternEndDate = EMPTY_WSTRING;
    }

    // for compatibility
    if (noEndDate) {
        occurrences = 0;
    }

    // CHANGE-DAY to UTC: modify monthOfYear / dayOfMonth / dayOfWeekMask.
    // All-day event and tasks don't need conversions.
    if ( USE_CHANGE_DAY 
         && (start != EMPTY_WSTRING)
         && !(isSystemTimeDate(start)) 
         && USE_UTC ) {

        changeDay(L"UTC");
    }


    //
    // Read all exceptions.
    //
    appExceptions.clear();
    numRecursions = 0;
    int numExceptions = getExceptionsCountOnClient();
    ClientAppException* exc = NULL;
    for (int i=0; i<numExceptions; i++) {
        exc = getExceptionOnClient(i);
        if (exc) {
            appExceptions.push_back(*exc);
            delete exc;
            exc = NULL;
        }
    }


    isUpdated = true;
    return 0;
}



/*
 * Save all the properties to Outlook.
 * Properties are verified to be consistent all together, 
 * as Outlook doesn't accept wrong values.
 * If UTC is used, all props are converted to correct values.
 * 
 * @return: 0 if no errors
 */
int ClientRecurrence::save() {

    HRESULT hr;
    DATE patternStart, patternEnd;
    int i = -1;    // index for 'recurrenceProps' array.

    // Check COM Ptr / Recurrence active
    if (!pRec || !isRecurring()) {
        sprintf(lastErrorMsg, ERR_OUTLOOK_REC_NOT_SET, L"save");
        throwClientException(lastErrorMsg);
        return 1;
    }

    // This is required correct!
    if (recurrenceType < 0 || recurrenceType > 6) {
        i=0;
        goto error;
    }


    // TBD: check consistance of props for this recType...
    checkIfRecIsCorrect();


    //
    // --------- Conversions ---------
    //
    if (interval < 1) {
        interval = 1;
    }

    //
    // PatternStartDate: get value from <Start> if not empty.
    // (ignore the passed value and overwrite it with start - new from 6.0.4)
    if (start != EMPTY_WSTRING) {
        patternStartDate = start;
    }
    if (isSystemTimeDate(patternStartDate)) {
        systemTimeDateToDouble(patternStartDate, &patternStart);        // All-day -> no conversion is necessary.
    }
    else {
        systemTimeToDouble(patternStartDate, &patternStart, L"0000");   // UTC     -> convert to local time and set time to 00:00.
    }                                                                   //            (MUST put 00:00 to Outlook!)

    //
    // PatternEndDate: UTC value is expected, but accept also "yyyy-MM-dd"
    //
    if (isSystemTimeDate(patternEndDate)) {
        systemTimeDateToDouble(patternEndDate, &patternEnd);            // All-day -> no conversion is necessary.
    }
    else {
        systemTimeToDouble(patternEndDate, &patternEnd);                // UTC     -> convert to local time.
    }



    // CHANGE-DAY to Local: modify monthOfYear / dayOfMonth / dayOfWeekMask.
    // All-day event and tasks don't need conversions.
    if ( USE_CHANGE_DAY 
         && (start != EMPTY_WSTRING)
         && !(isSystemTimeDate(start)) 
         && USE_UTC ) {

        changeDay(L"Local");
    }


    //
    // -----------Put ALL ------------
    // (note that this order is mandatory to put props)
    //
    i=0;
    hr = pRec->put_RecurrenceType((OlRecurrenceType)recurrenceType);
    if (FAILED(hr)) goto error;
    
    i++;
    hr = pRec->put_Interval((long)interval);
    if (FAILED(hr)) goto error;
    
    i++; 
    // Ignore if '0' (default) or '-1' (tag not found)
    if (monthOfYear != 0 && monthOfYear != -1) {
        hr = pRec->put_MonthOfYear((long)monthOfYear);
        if (FAILED(hr)) goto error;
    }
    
    i++;
    // Ignore if '0' (default) or '-1' (tag not found)
    if (dayOfMonth != 0 && dayOfMonth != -1) {
        hr = pRec->put_DayOfMonth((long)dayOfMonth);
        if (FAILED(hr)) goto error;
    }
    
    i++;
    // Ignore if '0' (default) or '-1' (tag not found)
    if (dayOfWeekMask != 0 && dayOfWeekMask != -1) {
        hr = pRec->put_DayOfWeekMask((OlDaysOfWeek)dayOfWeekMask);
        if (FAILED(hr)) goto error;
    }
    
    i++;
    // Ignore if '0' (default) or '-1' (tag not found)
    if (instance != 0 && instance != -1) {
        hr = pRec->put_Instance((long)instance);
        if (FAILED(hr)) goto error;
    }
    
    i++;
    if (patternStart < LIMIT_MAX_DATE) {
        // Outlook wants PatternStartDate with time = "00:00".
        patternStart = (int)patternStart;
        hr = pRec->put_PatternStartDate(patternStart);
        if (FAILED(hr)) goto error;
    }
  
    i++;
    hr = pRec->put_NoEndDate(BOOLToVBool(noEndDate));
    if (FAILED(hr)) goto error;


    i++;
    // Ignore if '0' (default) or '-1' (tag not found)
    // Note: set "patternEndDate" only if "occurrences" not received or wrong.
    // ----- (ignore patternEndDate if values disagree)
    if (occurrences != 0 && occurrences != -1) {
        i++;
        hr = pRec->put_Occurrences((long)occurrences);
        if (FAILED(hr)) goto error;
    }
    else {
        if (patternEnd < LIMIT_MAX_DATE) {
            hr = pRec->put_PatternEndDate(patternEnd);
            if (FAILED(hr)) goto error;
        }
        i++;
    }

    isUpdated = true;
    return 0;

error:
    isUpdated = false;
    if (i>=0 && recurrenceProps[i]) {
        sprintf(lastErrorMsg, ERR_OUTLOOK_REC_SAVE, recurrenceProps[i]);
        LOG.error(lastErrorMsg);
        throwClientException(lastErrorMsg);
    }
    return 1;
}
///////////////////////////////////////////////////////////////////////////



void ClientRecurrence::checkIfRecIsCorrect() {
    //
    // TBD: verify if some props are not correct for the rec type.
    // (display warning?)
    // http://msdn.microsoft.com/library/default.asp?url=/library/en-us/vbaol11/html/olobjRecurrencePattern_HV05247371.asp
    //
}



/*
 * Change-day option: if USE_CHANGE_DAY is true, rec pattern props
 * are changed according to Start value and USE_UTC.
 * Returns true if change-day is applied
 */
bool ClientRecurrence::changeDay(const wstring dest) {
    
    DATE date = 0;
    SYSTEMTIME time, localTime, utcTime;

    // This call gets 'date' ALWAYS in local time...
    systemTimeToDouble     (start, &date);
    VariantTimeToSystemTime(date,  &time);

    localTime = time;
    localTimeToUTC(time);
    utcTime   = time;


    // Day to be changed!
    if (localTime.wDay != utcTime.wDay) {
        if (dest == L"Local") {
            time = localTime;
        }

        //
        // <MonthOfYear>: get from <Start>
        //
        if (monthOfYear != 0 && monthOfYear != -1) {
            monthOfYear = time.wMonth;
        }

        //
        // <DayOfMonth>: get from <Start>
        //
        if (dayOfMonth != 0 && dayOfMonth != -1) {
            dayOfMonth = time.wDay;
        }

        //
        // <DayOfWeekMask>: shift the mask
        //
        if (dayOfWeekMask != 0 && dayOfWeekMask != -1) {
            
            // Calculate if forward or backward 1 day.
            TIME_ZONE_INFORMATION tmz;
            GetTimeZoneInformation(&tmz);
            bool forward;
            if (dest == L"Local") {
                if (tmz.Bias < 0) forward = true;
                else              forward = false;
            }
            else {
                if (tmz.Bias < 0) forward = false;
                else              forward = true;
            }

            // Forward = left shift.
            if (forward) {
                dayOfWeekMask = dayOfWeekMask << 1;
                if (dayOfWeekMask > 127) {
                    dayOfWeekMask -= 127;
                }
            }
            // Backward = right shift.
            else {
                if (dayOfWeekMask & 1) {
                    dayOfWeekMask += 127;
                }
                dayOfWeekMask = dayOfWeekMask >> 1;
            }
        }
        return true;
    }
    return false;
}



/*
 * Methods to get/set internal members.
 * -------------------------------------------------------
 */
const int ClientRecurrence::getRecurrenceType() {
    if (!isUpdated) {
        read();
    }
    return recurrenceType;
}
void ClientRecurrence::setRecurrenceType(const int val) {
    recurrenceType = val;
    isUpdated = false;
}

const int ClientRecurrence::getInterval() {
    if (!isUpdated) {
        read();
    }
    return interval;
}
void ClientRecurrence::setInterval(const int val) {
    interval = val;
    isUpdated = false;
}

const int ClientRecurrence::getMonthOfYear() {
    if (!isUpdated) {
        read();
    }
    return monthOfYear;
}
void ClientRecurrence::setMonthOfYear(const int val) {
    monthOfYear = val;
    isUpdated = false;
}

const int ClientRecurrence::getDayOfMonth() {
    if (!isUpdated) {
        read();
    }
    return dayOfMonth;
}
void ClientRecurrence::setDayOfMonth(const int val) {
    dayOfMonth = val;
    isUpdated = false;
}

const int ClientRecurrence::getDayOfWeekMask() {
    if (!isUpdated) {
        read();
    }
    return dayOfWeekMask;
}
void ClientRecurrence::setDayOfWeekMask(const int val) {
    dayOfWeekMask = val;
    isUpdated = false;
}

const int ClientRecurrence::getInstance() {
    if (!isUpdated) {
        read();
    }
    return instance;
}
void ClientRecurrence::setInstance(const int val) {
    instance = val;
    isUpdated = false;
}

const wstring& ClientRecurrence::getPatternStartDate() {
    if (!isUpdated) {
        read();
    }
    return patternStartDate;
}
void ClientRecurrence::setPatternStartDate(const wstring& val) {
    patternStartDate = val;
    isUpdated = false;
}

const BOOL ClientRecurrence::getNoEndDate() {
    if (!isUpdated) {
        read();
    }
    return noEndDate;
}
void ClientRecurrence::setNoEndDate(const BOOL val) {
    noEndDate = val;
    isUpdated = false;
}

const wstring& ClientRecurrence::getPatternEndDate() {
    if (!isUpdated) {
        read();
    }
    return patternEndDate;
}
void ClientRecurrence::setPatternEndDate(const wstring& val) {
    patternEndDate = val;
    isUpdated = false;
}

const int ClientRecurrence::getOccurrences() {
    if (!isUpdated) {
        read();
    }
    return occurrences;
}
void ClientRecurrence::setOccurrences(const int val) {
    occurrences = val;
    isUpdated = false;
}


void ClientRecurrence::setStart(const wstring& val) {
    start = val;
}




// ------------------------- Methods to manage Appointment Exceptions -----------------------------


/**
 * Returns the number of exceptions stored in this recurrence.
 */
const int ClientRecurrence::getExceptionsCount() {
    if (!isUpdated) {
        read();
    }
    return appExceptions.size();
}


/**
 * Returns the desired ClientAppException stored in 'appExceptions' internal list.
 * @param index : the index of exception desired
 * @return      : the ClientAppException pointer (to internal item, MUST NOT free it)
 */
ClientAppException* ClientRecurrence::getException(const int index) {

    if (index<0 || (unsigned int)index>=appExceptions.size()) {
        return NULL;
    }

    exceptionIterator it = appExceptions.begin();
    for (int i=0; i<index; i++) {
        it++;
    }
    return &(*it);
}


/**
 * Adds a ClientAppException into 'appExceptions' internal list.
 * The object passed is copied, so can be safely freed by the caller.
 * @param cException : the ClientAppException* to add
 * @return           : the number of exceptions stored after the add operation
 */
int ClientRecurrence::addException(ClientAppException* cException) {

    if (cException) {
        appExceptions.push_back(*cException);
    }
    return appExceptions.size();
}


/**
 * Reset the exception list 'appExceptions'.
 */
void ClientRecurrence::resetExceptions() {
    appExceptions.clear();
}



/**
 * This utility method is used to remove all duplicated exceptions that are
 * extracted from Outlook.
 * We don't want more than one exception with the same "OriginalDate" (it's nonsense...).
 * Outlook may send more than one, for example:
 *
 *  <Exception>
 *    <OriginalDate>2007-01-15</OriginalDate>                                              
 *    <ExAppointment>
 *        <ExSubject>Subject modified</ExSubject>          
 *        <ExBody/>         
 *        <ExLocation/>  
 *        <ExStart>2007-01-16</ExStart>          
 *        <ExEnd>2007-01-16</ExEnd>      
 *        <ExAllDayEvent>1</ExAllDayEvent>
 *        <ExBusyStatus>1</ExBusyStatus> 
 *    </ExAppointment>    
 *  </Exception>
 *  <Exception>
 *    <OriginalDate>2007-01-15</OriginalDate>  
 *  </Exception>
 *
 * In this case, the second exception is nonsense, and will be removed from list.
 * 
 * @return : the number of exceptions removed from list.
 */
int ClientRecurrence::removeDuplicatedExceptions() {

    if (!appExceptions.size()) {
        return 0;
    }

    bool restart = false;
    int numRemoved = 0;
    exceptionIterator i = appExceptions.begin();

    //
    // Cycle through all exceptions
    //
    while (i != appExceptions.end()) {
        // Jump normal exceptions
        if ( (*i).getDeleted() == FALSE ) {
            i++;
            continue;
        }

        // Search for another exception with the same "OriginalDate"
        exceptionIterator j = i;
        j++;
        while (j != appExceptions.end()) {
            if ( (int)((*i).getOriginalDate()) == (int)((*j).getOriginalDate()) ) {
                // Found! -> remove the one wich is a deleted occurrence.
                appExceptions.erase(i);
                numRemoved ++;
                restart = true;
                break;
            }
            j++;
        }

        if (restart) {
            // Need to restart again, "i" is not valid now.
            restart = false;
            i = appExceptions.begin();
            continue;
        }
        i++;
    }

    return numRemoved;
}





/**
 * Returns the number of appointment exceptions.
 * Value is always retrieved from Outlook (not saved in this object).
 * This method is used from client to server, initially to populate
 * the exceptions list.
 */
const int ClientRecurrence::getExceptionsCountOnClient() {

    // Not recurring -> no exceptions :)
    if (!isRecurring()) {
        return 0;
    }

    int numExceptions = 0;
    try {
        ExceptionsPtr pExceptions = pRec->GetExceptions();
        if (!pExceptions) return 0;
        numExceptions = pExceptions->GetCount();
    }
    catch (_com_error &e) {
        manageComErrors(e);
        return 0;
    }

    return numExceptions;
}



/**
 * Returns the ClientAppException object of the desired index [0, numExceptions-1].
 * Value is always retrieved from Outlook (not saved in this object).
 * This method is used from client to server, initially to populate
 * the exceptions list.
 * Note: the pointer returned is a new allocated object.
 *
 * @param index : the index of Exception required
 * @return      : the correspondent pointer (new allocated) of ClientAppException (NULL if not found).
 */
ClientAppException* ClientRecurrence::getExceptionOnClient(const int index) {

    // Not recurring -> no exceptions :)
    if (!isRecurring()) {
        return NULL;
    }

    ClientAppException* appException = NULL;

    try {
        ExceptionsPtr pExceptions = pRec->GetExceptions();
        if (!pExceptions) return NULL;
        int numExceptions = pExceptions->GetCount();

        if (index<0 || index>=numExceptions) return NULL;

        // Get the appointment exception.
        ExceptionPtr pException = pExceptions->Item(index+1);       // First index = 1.
        if (!pException) return NULL;

        appException = new ClientAppException();
        appException->setCOMPtr(pException);
    }

    catch (_com_error &e) {
        manageComErrors(e);
        return NULL;
    }

    return appException;
}



/**
 * Returns the specific occurrence of date "originalDate" from a recurring appointment.
 * If the occurrence is not found, returns NULL.
 *
 * @param originalDate : the original date of the occurrence to be found
 *                       (double format)
 * @return             : the new COM pointer "_AppointmentItemPtr" which is the desired occurrence
 *                       (NULL if not found)
 */
_AppointmentItemPtr ClientRecurrence::getOccurrence(const DATE originalDate) {

    if (!isRecurring()) {
        return NULL;
    }
    _AppointmentItemPtr pOcc = NULL;

    try {
        pOcc = pRec->GetOccurrence(originalDate);
        if (!pOcc) {
            return NULL;        // Not found
        }
    }
    catch (_com_error&) {
        return NULL;            // Not found
    }

    return pOcc;
}



/**
 * Save all exceptions from internal exception list to Outlook (server to client).
 * First we save the 'deleted occurrence', that are easier.
 * Then all other exceptions are saved using method 'safeSaveException': 
 * we MUST save exceptions in the right order, there can be dependences between
 * exceptions (if the "StartDate" of an exceptions e1 is equal to the "OriginalDate"
 * of an exception e2, we MUST first execute the exception e2).
 *
 * @return : 0 if all exceptions saved correctly
 */
int ClientRecurrence::saveAllExceptions() {

    int ret = 0;
    int numExceptions = appExceptions.size();
    if (numExceptions>0) {

        // First save all exceptions which are deleted occurrence.
        exceptionIterator it = appExceptions.begin();
        while (it != appExceptions.end()) {
            if ( (*it).getDeleted() ) {
                saveException(&(*it));
            }
            it++;
        }

        //
        // ****** SAVE OF MODIFIED EXCEPTIONS IS DISABLED ******
        // Only deleted exceptions are currently used, modified exc
        // are transformed into separate events.
        //
        // Then save all other exceptions (safeSave).
        //numRecursions = 0;
        //it = appExceptions.begin();
        //while (it != appExceptions.end()) {
        //    ret += safeSaveException(&(*it));
        //    it++;
        //}
    }
    return ret;
}


//
// ****** SAVE OF MODIFIED EXCEPTIONS IS DISABLED ******
// Only deleted exceptions are currently used, modified exc
// are transformed into separate events.
//
/**
 * Save the exception 'cException' into Outlook.
 * First check if this exception depends on another exception of the list.
 * In this case, we first execute that exception (recursive call).
 * Than saves this exception to Outlook.
 *
 * @return 0 if no errors
 */
//int ClientRecurrence::safeSaveException(ClientAppException* cException) {
//
//    // Already saved -> nothing to do.
//    if (cException->isSaved()) {
//        return 0;
//    }
//    // Occurrence to delete -> save without problem
//    if (cException->getDeleted()) {
//        return saveException(cException);
//    }
//
//    //
//    // OriginalDate = StartDate -> save without problem
//    //
//    int thisOriginalDay = (int)cException->getOriginalDate();
//    wstring thisStart = cException->getStart();
//    DATE thisStartDate = 0;
//    if (thisStart.size() > 0) {
//        systemTimeDateToDouble(thisStart, &thisStartDate);
//    }
//    if (thisOriginalDay == (int)thisStartDate) {
//        return saveException(cException);
//    }
//
//    
//    // Check to avoid recursion deadlock...
//    if (numRecursions > ((int)appExceptions.size() + 1)) {
//        LOG.error(ERR_EXCEPTIONS_DEADLOCK, numRecursions);
//        return 1;
//    }
//
//    //
//    // Search if another exception has the OriginalDate = this startDate.
//    // In this case there's a dependence: we first save that exception.
//    //
//    exceptionIterator it = appExceptions.begin();
//    while (it != appExceptions.end()) {
//
//        if ( (*it).getOriginalDate() == cException->getOriginalDate() ){
//            it++;      // It's me...
//            continue;
//        }
//
//        if ((int)thisStartDate == (int)(*it).getOriginalDate()) {
//            // *** Save first that exception (recursive call!) ***
//            numRecursions++;
//            safeSaveException(&(*it));
//        }
//        it++;
//    }
//
//    // Finally save this exception
//    return saveException(cException);
//}



/**
 * Save the exception 'cException' into Outlook.
 * No check on exceptions dependences is done here.
 * Before saving the exception, we MUST check that the destination date
 * is free (the "StartDate") -> call 'freeDestinationDays()'.
 * Note: before calling this method, the Appointment Item MUST be
 *       already saved to Outlook.
 *
 * @return 0 if no errors
 */
int ClientRecurrence::saveException(ClientAppException* cException) {

    //
    // Get the occurrence (appointment item MUST be already saved)
    //
    DATE originalDate = cException->getOriginalDate();
    if (!originalDate) { 
        // Error: no originalDate
        sprintf(lastErrorMsg, ERR_PROPERTY_REQUIRED, L"appointment exception", L"OriginalDate");
        LOG.error(lastErrorMsg);
        return 1;
    }

    // *** MUST correct the originalDate in case of ***
    // - occurrence deleted  (yes, only deleted now supported)
    // - not all day event
    if (!isSystemTimeDate(start)) {
        wstring realOriginalDate = cException->formatOriginalDate(FALSE, start);
        systemTimeToDouble(realOriginalDate, &originalDate);
    }

    _AppointmentItemPtr pOcc = getOccurrence(originalDate);
    if (!pOcc) {
        // Occurrence not found - simply ignore.
        // sprintf(lastErrorMsg, ERR_OCCURRENCE_NOT_FOUND, cException->formatOriginalDate().c_str());
        // LOG.debug(lastErrorMsg);
        return 1;
    }

    //
    // Link the appointment occurrence pointer
    //
    cException->setCOMPtr(pOcc);


    //
    // Save exception / delete occurrence.
    //
    if (cException->getDeleted()) {
        if (cException->deleteOccurrence()) {
            sprintf(lastErrorMsg, ERR_OCCURRENCE_NOT_DELETED, cException->formatOriginalDate().c_str());
            LOG.error(lastErrorMsg);
            return 1;
        }
    }
    else {
        sprintf(lastErrorMsg, "Internal error: only 'deleted' exceptions can be saved. Trying to save a modified exception (date = %ls).", cException->formatOriginalDate().c_str());
        LOG.debug(lastErrorMsg);
        return 1;
        //
        // ****** SAVE OF MODIFIED EXCEPTIONS IS DISABLED ******
        // Only deleted exceptions are currently used, modified exc
        // are transformed into separate events.
        //
        // Need to free the destination day (and all days between "OriginalDate" and "Start")
        //DATE destination;
        //systemTimeDateToDouble(cException->getStart(), &destination);
        //freeDestinationDays(destination, cException->getOriginalDate());

        //if (cException->saveOccurrence()) {
        //    sprintf(lastErrorMsg, ERR_OCCURRENCE_NOT_SAVED, cException->formatOriginalDate().c_str());
        //    LOG.error(lastErrorMsg);
        //    return 1;
        //}
    }

    return 0;
}


//
// ****** SAVE OF MODIFIED EXCEPTIONS IS DISABLED ******
// Only deleted exceptions are currently used, modified exc
// are transformed into separate events.
// This utility method is currently not used.
//
/**
 * Delete all occurrences between "OriginalDate" and "StartDate" (if found).
 * We check all days in this interval.
 */
//void ClientRecurrence::freeDestinationDays(const DATE startDate, const DATE originalDate) {
//
//    int originalDay = (int)originalDate;
//    int startDay    = (int)startDate;
//    DATE date1, date2;
//    _AppointmentItemPtr pOcc = NULL;
//
//    // Same day: nothing to do
//    if (originalDay == startDay) {
//        return;
//    }
//
//    // Get date interval [date1 - date2] of occurrences to delete.
//    if (originalDay < startDay) {
//        date1 = originalDate + 1;
//        date2 = startDate;
//    }
//    else {
//        date1 = startDate;
//        date2 = originalDate - 1;
//    }
//
//    
//    //
//    // Delete all occurrences found between date1 and date2.
//    //
//    for (DATE i=date1; i<=date2; i++) {
//        pOcc = getOccurrence(i);
//        if (!pOcc) continue;
//
//        try {
//            HRESULT hr = pOcc->Delete();
//            if (FAILED(hr)) {
//                wstring tmp = EMPTY_WSTRING;
//                doubleToSystemTimeDate(tmp, i);
//                sprintf(lastErrorMsg, ERR_OCCURRENCE_NOT_DELETED, tmp);
//                LOG.error(lastErrorMsg);
//                continue;
//            }
//        }
//        catch(_com_error &e) {
//            manageComErrors(e);
//            wstring tmp = EMPTY_WSTRING;
//            doubleToSystemTimeDate(tmp, i);
//            sprintf(lastErrorMsg, ERR_OCCURRENCE_NOT_DELETED, tmp);
//            LOG.error(lastErrorMsg);
//            continue;
//        }
//    }
//}





//
// ---------------- default ones are OK ------------------
//
///* 
// * Copy Constructor
// */
//ClientRecurrence::ClientRecurrence(const ClientRecurrence& c) {
//
//    pRec             = c.pRec;
//    recurring        = c.recurring;
//    isUpdated        = c.isUpdated;
//
//    recurrenceType   = c.recurrenceType;
//    interval         = c.interval;
//    monthOfYear      = c.monthOfYear;
//    dayOfMonth       = c.dayOfMonth;
//    dayOfWeekMask    = c.dayOfWeekMask;
//    instance         = c.instance;
//    patternStartDate = c.patternStartDate;
//    noEndDate        = c.noEndDate;
//    patternEndDate   = c.patternEndDate;
//    occurrences      = c.occurrences;
//}
//
//
///* 
// * Operator =
//*/
//ClientRecurrence ClientRecurrence::operator=(const ClientRecurrence& c) {
//
//    ClientRecurrence cnew;
//
//    cnew.pRec             = c.pRec;
//    cnew.recurring        = c.recurring;
//    cnew.isUpdated        = c.isUpdated;
//
//    cnew.recurrenceType   = c.recurrenceType;
//    cnew.interval         = c.interval;
//    cnew.monthOfYear      = c.monthOfYear;
//    cnew.dayOfMonth       = c.dayOfMonth;
//    cnew.dayOfWeekMask    = c.dayOfWeekMask;
//    cnew.instance         = c.instance;
//    cnew.patternStartDate = c.patternStartDate;
//    cnew.noEndDate        = c.noEndDate;
//    cnew.patternEndDate   = c.patternEndDate;
//    cnew.occurrences      = c.occurrences;
//
//    return cnew;
//}

