/*
 * Copyright (C) 2005-2006 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "outlook/defs.h"
#include "outlook/ClientException.h"
#include "base/Log.h"
#include "winmaincpp.h"
#include "outlook/utils.h"
#include "outlook/itemProps.h"

#include <algorithm>
#include <string>

using namespace std;



//
// -------------- Conversion functions for item type coding ---------------------
//

/*
 * itemType (string) -> OlDefaultFolders (int)
 */
const OlDefaultFolders getDefaultFolderType(const wstring& itemType) {

    for(int i=0; i<ITEM_TYPES_COUNT; i++) {
        if (itemTypes[i].name == itemType) {
            return itemTypes[i].olFolderType;
        }
    }

    // If not found
    sprintf(lastErrorMsg, ERR_OUTLOOK_BAD_ITEMTYPE, itemType.c_str());
    throwClientFatalException(lastErrorMsg);
    return (OlDefaultFolders)NULL;
}


/*
 * itemType (string) -> OlItemType (int)
 */
const OlItemType getOlItemType(const wstring& itemType) {

    for(int i=0; i<ITEM_TYPES_COUNT; i++) {
        if (itemTypes[i].name == itemType) {
            return itemTypes[i].olType;
        }
    }

    // If not found
    sprintf(lastErrorMsg, ERR_OUTLOOK_BAD_ITEMTYPE, itemType.c_str());
    throwClientFatalException(lastErrorMsg);
    return (OlItemType)NULL;
}


/*
 * OlItemType (int) -> itemType (string)
 */
const wstring getItemTypeFromOlType(const OlItemType olType) {

    for(int i=0; i<ITEM_TYPES_COUNT; i++) {
        if (itemTypes[i].olType == olType) {
            return itemTypes[i].name;
        }
    }

    // If not found
    sprintf(lastErrorMsg, ERR_OUTLOOK_BAD_OLTYPE, olType);
    throwClientFatalException(lastErrorMsg);
    return NULL;
}
// ----------------------------------------------------------------------------



/*
 * Actions to execute when a COM pointer exception occurs.
 */
void manageComErrors(_com_error &e) {

    sprintf(lastErrorMsg, ERR_COM_POINTER, e.Error(), e.ErrorMessage());
    lastErrorCode = (int)e.Error(); 

    LOG.error(lastErrorMsg);
    //safeMessageBox(lastErrorMsg);
}





/*
 * Return the program name from version number.
 *
 * 12  -> Outlook 2007  *** TBD: NOT TESTED ***
 * 11  -> Outlook 2003
 * 10  -> Outlook XP
 * 9   -> Outlook 2000
 * 8.5 -> Outlook 98    *** TBD: NOT TESTED ***
 * 8.0 -> Outlook 97    *** TBD: NOT TESTED ***
 */
const wstring getNameFromVersion(wstring version) {

    int majorVersion = _wtoi(version.c_str());
    wstring name = EMPTY_WSTRING;

    switch (majorVersion) {
        case 12: {
            name = L"Outlook 2007";
            break;
		}
        case 11: {
            name = L"Outlook 2003";
            break;
		}
        case 10: {
            name = L"Outlook XP";
            break;
		}  
        case 9: {
            name = L"Outlook 2000";
            break;
		}                                                     
        case 8: {
            int minorVersion = _wtoi(&version[2]);
            if (minorVersion == 0) {
                name = L"Outlook 97";
            }
            else if (minorVersion == 5) {
                name = L"Outlook 98";
            }
            else {
                goto error;
            }
            break;
		}                                                  
		default: {
			goto error;
		}
	}

    return name;

error:
    sprintf(lastErrorMsg, ERR_OUTLOOK_INVALID_VERSION, version.c_str());
    throwClientFatalException(lastErrorMsg);
    return NULL;
}


/**
 * Converts a string to lower-case.
 * String passed is modified.
 * @param s   [IN-OUT] the string to be converted
 */
void toLowerCase(string& s) {
    transform(s.begin(), s.end(), s.begin(), tolower);
}

/**
 * Converts a string to lower-case (wstring version).
 * String passed is modified.
 * @param s   [IN-OUT] the string to be converted
 */
void toLowerCase(wstring& s) {
    transform(s.begin(), s.end(), s.begin(), tolower);
}



//
// ------------------------------ DATE/TIME CONVERSIONS FUNCTIONS ------------------------------
//
/*
 * Variant time (double) -> System time ("YYYYMMDDThhmmss" or "YYYYMMDDThhmmssZ")
 * The output value is a string. It's in UTC based on param 'toUTC'.
 *
 * @param systemDate   : [OUT] the date returned in SystemTime format
 * @param doubleDate   : the input date in variant time format
 * @param toUTC        : TRUE if we want conversion in UTC format (YYYYMMDDThhmmssZ)
 * @param setHourToZero: (optional) if TRUE set time to 00:00 (default = FALSE)
 */
void doubleToSystemTime(wstring& systemDate, const DATE doubleDate, const BOOL toUTC, const BOOL setHourToZero) {
    
    if (doubleDate > LIMIT_MAX_DATE) {
        systemDate = EMPTY_WSTRING;
        return;
    }

    SYSTEMTIME t;
    VariantTimeToSystemTime(doubleDate, &t);

    if (setHourToZero) {
        t.wHour = 0;
        t.wMinute = 0;
        t.wSecond = 0;
    }
    if (toUTC) {
        localTimeToUTC(t);
    }

    WCHAR date  [20];
    WCHAR month [10];
    WCHAR day   [10];
    WCHAR hour  [10];
    WCHAR minute[10];
    WCHAR second[10];

    // month
    if (t.wMonth < 10)
        wsprintf(month, TEXT("0%i"), t.wMonth);
    else
        wsprintf(month, TEXT("%i"), t.wMonth);

    // day
    if (t.wDay  < 10)
        wsprintf(day, TEXT("0%i"), t.wDay);
    else
        wsprintf(day, TEXT("%i"), t.wDay);

    // hour
    if (t.wHour  < 10)
        wsprintf(hour, TEXT("0%i"), t.wHour);
    else
        wsprintf(hour, TEXT("%i"), t.wHour);

    // minute
    if (t.wMinute < 10)
        wsprintf(minute, TEXT("0%i"), t.wMinute);
    else
        wsprintf(minute, TEXT("%i"), t.wMinute);

    // second
    if (t.wSecond < 10)
        wsprintf(second, TEXT("0%i"), t.wSecond);
    else
        wsprintf(second, TEXT("%i"), t.wSecond);

    if (toUTC) {
        wsprintf(date, TEXT("%i%s%sT%s%s%sZ"), t.wYear, month, day, hour, minute, second);
    } else {
        wsprintf(date, TEXT("%i%s%sT%s%s%s"), t.wYear, month, day, hour, minute, second);
    }
    
    systemDate = date;
}



/*
 * Variant time (double) -> System time ("yyyy-MM-dd")
 * It used for anniversary and birthday, where only date is required.
 * UTC is ignored (only date, no time).
 *
 * @param systemDate: [OUT] wstring with the date in yyyy-MM-dd format
 * @param doubleDate: the date in double format
 */
void doubleToSystemTimeDate(wstring& systemDate, const DATE doubleDate) {

    if (doubleDate > LIMIT_MAX_DATE) {
        systemDate = EMPTY_WSTRING;
        return;
    }

    SYSTEMTIME t;
    WCHAR month [10];
    WCHAR day   [10];
    WCHAR date  [11];

    VariantTimeToSystemTime(doubleDate, &t);  
    t.wHour = 0;
    t.wMinute = 0;
    t.wSecond = 0;

    // month
    if (t.wMonth < 10)
        wsprintf(month, L"0%i", t.wMonth);
    else
        wsprintf(month, L"%i", t.wMonth);
    
    // day
    if (t.wDay  < 10)
        wsprintf(day, L"0%i", t.wDay);
    else
        wsprintf(day, L"%i", t.wDay);
    

    wsprintf(date, L"%i-%s-%s", t.wYear, month, day);
    systemDate = date;
}



/*
 * Sysyem time ("yyyy-MM-dd") -> Variant time (double).
 * If empty string passed, returns REFERRED_MAX_DATE.
 * UTC is ignored (only date, no time).
 *
 * @param dataString : the input string in System time format
 * @param date       : [OUT] the returned value into VariantTime format
 * @param type       : used to set the hours and minutes:
 *                     "END"   to set hour 23:59
 *                     "START" to set hour 00:00 (default)
 * 
 * Set the hour to 00:00 is used for birthday, anniversary and 
 * startDate of appointment when it is all day event. 
 * Set hour 23:59 for endDate when an appointment is allDayEvent.
 */
void systemTimeDateToDouble(const wstring& dataString, DATE* date, WCHAR* type) {

    WCHAR* token  = NULL;
    WCHAR* dummy  = NULL;
    WCHAR* Ztoken = NULL;
    WCHAR tmp[50];
    int ret = 0;
    SYSTEMTIME t;

    WCHAR internalString[20];
    wsprintf(internalString, dataString.c_str());

    Ztoken = wcschr(internalString, '-');
    if (Ztoken == NULL) {
        // Note: DO NOT REMOVE THIS!
        // If wrong format ("YYYYMMDDThhmmssZ") the correct function is called...
        systemTimeToDouble(dataString, date, NULL);
        return;
    }

    //  TEXT("yyyy-MM-dd")
    token = wcstok(internalString, L"-");

    if (token != NULL) {
        //yyyy
        wcsncpy(tmp, internalString, 4);
        tmp[4] = 0;
        swscanf(tmp, L"%d",&t.wYear);
        //MM
        wcsncpy(tmp, internalString + 5, 2);
        tmp[2] = 0;
        swscanf(tmp, L"%d",&t.wMonth);
        //dd
         wcsncpy(tmp, internalString + 8, 2);
        tmp[2] = 0;
        swscanf(tmp, L"%d",&t.wDay);

        if (wcscmp(type, L"END") == 0) {
            t.wHour = 23;
            t.wMinute = 59;
            t.wSecond = 0;
        } 
        else {
            t.wHour = 0;
            t.wMinute = 0;
            t.wSecond = 0;
        }

        // set to default
        t.wMilliseconds = 0;
        t.wDayOfWeek = 0;
        ret = SystemTimeToVariantTime(&t, date);
    } 
    else {
        (*date) = REFERRED_MAX_DATE;
    }
}


/*
 * System time ("YYYYMMDDThhmmss" or "YYYYMMDDThhmmssZ") -> Variant time (double).
 * Automatic conversion from UTC to Local time if input time is UTC.
 *
 * @param dataString : the input string in System time format
 * @param date       : [OUT] the returned value into VariantTime format
 * @action           : the parameter to set manually hour and minute. If NULL current value are used
 *                     - if = L"0000" -> set time to 00:00
 *                     - if = L"2359" -> set time to 23:59
 */
void systemTimeToDouble(const wstring& dataString, DATE* date, WCHAR* action) {

    WCHAR* token  = NULL;
    WCHAR* dummy  = NULL;
    WCHAR* Ztoken = NULL;
    BOOL isUTC    = FALSE;
    WCHAR inputTime[20];
    WCHAR tmp[50];
    int ret = 0;
    SYSTEMTIME t;
    
    wsprintf(inputTime, dataString.c_str());
    token = wcstok(inputTime, L"T");

    wsprintf(inputTime, dataString.c_str());
    Ztoken = wcschr(inputTime, 'Z');
    if (Ztoken) {
        isUTC = TRUE;
    }

    if (token != NULL) {
        //yyyy
        wcsncpy(tmp, inputTime, 4);
        tmp[4] = 0;
        swscanf(tmp, L"%d",&t.wYear);
        //MM
        wcsncpy(tmp, inputTime + 4, 2);
        tmp[2] = 0;
        swscanf(tmp, L"%d",&t.wMonth);
        //dd
         wcsncpy(tmp, inputTime + 6, 2);
        tmp[2] = 0;
        swscanf(tmp, L"%d",&t.wDay);


        if (wcslen(inputTime) > 9) {
            //HH
            wcsncpy(tmp, inputTime + 9, 2);
            tmp[2] = 0;
            swscanf(tmp, L"%d",&t.wHour);
            //mm
             wcsncpy(tmp, inputTime + 11, 2);
            tmp[2] = 0;
            swscanf(tmp, L"%d",&t.wMinute);
            //ss
             wcsncpy(tmp, inputTime + 13, 2);
            tmp[2] = 0;
            swscanf(tmp, L"%d",&t.wSecond);
        } 
        else {
            t.wHour = 0;
            t.wMinute = 0;
            t.wSecond = 0;
        }

        // set to default
        t.wMilliseconds = 0;
        t.wDayOfWeek = 0;

        // Always returns Local time
        if(isUTC) {
            UTCToLocalTime(t);
        }

        if (action != NULL && wcscmp(action, L"0000") == 0) {
            t.wHour = 0;
            t.wMinute = 0;
            t.wSecond = 0;
        } 
        else if (action != NULL && wcscmp(action, L"2359") == 0) {
            t.wHour = 23;
            t.wMinute = 59;
            t.wSecond = 59;
        } 

        ret = SystemTimeToVariantTime (&t, date);
    }

    else {
        (*date) = REFERRED_MAX_DATE;
    }
}


/*
 * Return true if date passed is in format "yyyy-MM-dd".
 */
bool isSystemTimeDate(const std::wstring& dataString) {

    WCHAR* Ztoken = NULL;
    WCHAR internalString[20];

    wsprintf(internalString, dataString.c_str());
    Ztoken = wcschr(internalString, '-');
    if (Ztoken) {
        return true;
    }
    else {
        return false;
    }
}


/**
 * Converts 'sysTime' from local time to UTC.
 * Uses the specific settings for the time zone and daylight saving time
 * that are applied to the 'sysTime' date.
 */
void localTimeToUTC(SYSTEMTIME &sysTime) {

    TIME_ZONE_INFORMATION timeZoneInfo;
    SYSTEMTIME utcTime;

    GetTimeZoneInformation(&timeZoneInfo);
    TzSpecificLocalTimeToSystemTime(&timeZoneInfo, &sysTime, &utcTime);
    sysTime = utcTime;
}


/**
 * Converts 'sysTime' from UTC to local time.
 * Uses the specific settings for the time zone and daylight saving time
 * that are applied to the 'sysTime' date.
 */
void UTCToLocalTime(SYSTEMTIME &sysTime) {
    
    TIME_ZONE_INFORMATION timeZoneInfo;
    SYSTEMTIME localTime;

    GetTimeZoneInformation(&timeZoneInfo);
    SystemTimeToTzSpecificLocalTime(&timeZoneInfo, &sysTime, &localTime);
    sysTime = localTime;
}






//
// ------------------------------ BOOLEAN CONVERSIONS FUNCTIONS ------------------------------
//

// Variant bool (-1/0) to BOOL (1/0).
BOOL vBoolToBOOL(VARIANT_BOOL vbool) {
    if (vbool==VARIANT_TRUE)
        return TRUE;
    else
        return FALSE;
}

// BOOL (1/0) to Variant bool (-1/0).
VARIANT_BOOL BOOLToVBool(BOOL b) {
    if (b==TRUE)
        return VARIANT_TRUE;
    else
        return VARIANT_FALSE;
}