/*
 * Copyright (C) 2003-2007 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <shlobj.h>                     // Used by SHGetFolderPath()
#include <direct.h>                     // Used by _wmkdir()
#include <errno.h>                      // Used by _get_errno()

#include "spdm/constants.h"
#include "spdm/DMTreeFactory.h"
#include "spds/DataTransformerFactory.h"
#include "spds/B64Encoder.h"
#include "spds/B64Decoder.h"
#include "spds/DESEncoder.h"
#include "spds/DESDecoder.h"

#include "outlook/ClientException.h"
#include "outlook/utils.h"
#include "winmaincpp.h"
#include "OutlookConfig.h"
#include "utils.h"


using namespace std;



static HWND g_hwndTimedOwner;
static BOOL g_bTimedOut;

/**
 *
 *  MessageBoxTimer
 *
 *      The timer callback function that posts the fake quit message.
 *      This function causes the message box to exit because the message box
 *      has determined that the application is exiting.
 *
 */
static void CALLBACK MsgBoxTimer(HWND hw, UINT uiMsg, UINT idEv, DWORD time) {
    g_bTimedOut = TRUE;
    if (g_hwndTimedOwner) {
        EnableWindow(g_hwndTimedOwner, TRUE);
    }
    PostQuitMessage(0);
}


/**
 *
 *  TimedMessageBox
 *
 *      The same as the standard MessageBox, except that TimedMessageBox
 *      also accepts a timeout. If the user does not respond within the
 *      specified timeout, the value 0 is returned instead of one of the
 *      ID* values.
 *
 */
int TimedMessageBox(HWND hwndOwner,
                    LPCTSTR pszMessage,
                    LPCTSTR pszTitle,
                    UINT flags,
                    DWORD dwTimeout) {

    UINT idTimer;
    int iResult;

    g_hwndTimedOwner = NULL;
    g_bTimedOut = FALSE;

    if (hwndOwner && IsWindowEnabled(hwndOwner)) {
        g_hwndTimedOwner = hwndOwner;
    }

    // Set a timer to dismiss the message box.
    idTimer = SetTimer(NULL, 0, dwTimeout, (TIMERPROC)MsgBoxTimer);

    iResult = MessageBox(hwndOwner, pszMessage, pszTitle, flags);

    // Finished with the timer.
    KillTimer(NULL, idTimer);


    // See if there is a WM_QUIT message in the queue if we timed out.
    // Eat the message so we do not quit the whole application.
    if (g_bTimedOut) {
        MSG msg;
        PeekMessage(&msg, NULL, WM_QUIT, WM_QUIT, PM_REMOVE);
        iResult = -1;
    }

    return iResult;
}





/**
 * Returns true if dataType is one of accepted SIF formats. (wstring version)
 */
bool isSIF(const wstring& dataType) {

    if ( dataType == L"text/x-s4j-sifc" ||
         dataType == L"text/x-s4j-sife" ||
         dataType == L"text/x-s4j-sift" ||
         dataType == L"text/x-s4j-sifn" ){
        return true;
    }
    else {
        return false;
    }
}

/**
 * Returns true if dataType is one of accepted SIF formats. (string version)
 */
bool isSIF(const string& dataType) {

    if ( dataType == "text/x-s4j-sifc" ||
         dataType == "text/x-s4j-sife" ||
         dataType == "text/x-s4j-sift" ||
         dataType == "text/x-s4j-sifn" ) {
        return true;
    }
    else {
        return false;
    }
}


/**
 * Returns true if dataType is one of accepted formats by Outlook plugin. (wstring version)
 */
bool isAcceptedDataType(const wstring& dataType) {

    if ( dataType == L"text/x-s4j-sifc"  ||
         dataType == L"text/x-s4j-sife"  ||
         dataType == L"text/x-s4j-sift"  ||
         dataType == L"text/x-s4j-sifn"  ||
         dataType == L"text/x-vcard"     ||
         dataType == L"text/x-vcalendar" ){
        return true;
    }
    else {
        return false;
    }
}

/**
 * Returns true if dataType is one of accepted formats by Outlook plugin. (string version)
 */
bool isAcceptedDataType(const string& dataType) {

    if ( dataType == "text/x-s4j-sifc"  ||
         dataType == "text/x-s4j-sife"  ||
         dataType == "text/x-s4j-sift"  ||
         dataType == "text/x-s4j-sifn"  || 
         dataType == "text/x-vcard"     ||
         dataType == "text/x-vcalendar" ) {
        return true;
    }
    else {
        return false;
    }
}




/**
 * Returns the syncmode name given its code.
 */
char* syncModeName(SyncMode code) {

	switch (code) {
		case SYNC_NONE:
			return "none";
			break;
		case SYNC_TWO_WAY:
        case SYNC_TWO_WAY_BY_SERVER:
			return "two-way";
			break;
		case SYNC_SLOW:
			return "slow";
			break;
		case SYNC_ONE_WAY_FROM_CLIENT:
        case SYNC_ONE_WAY_FROM_CLIENT_BY_SERVER:
			return "one-way-from-client";
			break;
		case SYNC_REFRESH_FROM_CLIENT:
        case SYNC_REFRESH_FROM_CLIENT_BY_SERVER:
			return "refresh-from-client";
			break;
		case SYNC_ONE_WAY_FROM_SERVER:
        case SYNC_ONE_WAY_FROM_SERVER_BY_SERVER:
			return "one-way-from-server";
			break;
		case SYNC_REFRESH_FROM_SERVER:
        case SYNC_REFRESH_FROM_SERVER_BY_SERVER:
			return "refresh-from-server";
			break;
		default:
			return NULL;
			break;
	}
}


/**
 * Returns true if the passed sync mode is a 'full sync mode'
 * (slow sync, refresh from client, refresh from server).
 */
bool isFullSyncMode(SyncMode mode) {

    if ( mode == SYNC_SLOW ||
         mode == SYNC_REFRESH_FROM_SERVER ||
         mode == SYNC_REFRESH_FROM_CLIENT ) {
        return true;
    }
    else {
        return false;
    }
}




/**
 * Convert the path in Windows format, changing the slashes in back-slashes.
 * @param str - [IN-OUT] the string to convert
 */
void toWindows(char* str) {
    int i=0;
    while (str[i]) {
        if (str[i] == '/') {
            str[i] = '\\';
        }
        i++;
    }
}


/*
 * Replace all strings 'source' with 'dest' inside string 'dataString'.
 * @param source     : the token to search
 * @param dest       : the token to replace
 * @param dataString : [IN-OUT] the string to work on
 */
void replaceAll(const wstring& source, const wstring& dest, wstring& dataString) {

    wstring::size_type pos = 0;
    pos = dataString.find(source, pos);

    while (pos != wstring::npos) {
        dataString.replace(pos, source.length(), dest);
        pos = dataString.find(source, pos+1);
    }
}




/*
 * Gets the content of <tag>...</tag> inside the string 'xml'.
 * Returns 0 if tag found, 1 if not found.
 *
 * @param xml     : the input string to search inside
 * @param tag     : the tag we are looking for
 * @param content : [OUT] the content of tag (string passed by ref)
 * @param pos     : (optional) the starting position to search inside 'xml' (default = 0)
 * @return        : 0 if tag found (even if empty content), 1 if errors
 */
int getElementContent(const wstring& xml, const wstring& tag, wstring& content, unsigned int pos) {

    wstring openTag, closeTag, emptyTag;
    wstring::size_type start, end;
    content = EMPTY_WSTRING;

    //
    // First try to find the empty element <tag/>
    //
    emptyTag  = L"<";
    emptyTag += tag;
    emptyTag += L"/>";

    start = xml.find(emptyTag, pos); 
    if (start != wstring::npos) {
        return 0;  // found!
    }

    //
    // Now search for the <tag>Content</tag>
    //
    openTag   = L"<";
    openTag  += tag;
    openTag  += L">";

    closeTag  = L"</";
    closeTag += tag;
    closeTag += L">";

    start = xml.find(openTag, pos); 
    if (start != wstring::npos) {
        start += openTag.length();

        end = xml.find(closeTag, start);
        if (end == wstring::npos) {
            goto not_found;
        }

        content = xml.substr(start, end-start);
    }
    
    else {
        goto not_found;
    }

    return 0;

not_found:

    return 1;
}



/*
 * Gets the content of <tag>...</tag> inside the string 'xml'.
 * Returns 0 if tag found, 1 if not found.
 *
 * @param xml     : the input string to search inside
 * @param tag     : the tag we are looking for
 * @param content : [OUT] the content of tag (string passed by ref)
 * @param pos     : the starting position to search inside 'xml'
 * @param start   : [OUT] the starting position of content found ('wstring::npos' if not found)
 * @param end     : [OUT] the ending position of content found ('wstring::npos' if not found)
 * @return        : 0 if tag found (even if empty content), 1 if errors
 */
int getElementContent(const wstring& xml, const wstring& tag, wstring& content, const wstring::size_type pos, wstring::size_type& start, wstring::size_type& end) {

    wstring openTag, closeTag, emptyTag;
    start   = wstring::npos;
    end     = wstring::npos;
    content = EMPTY_WSTRING;

    //
    // First try to find the empty element <tag/>
    //
    emptyTag  = L"<";
    emptyTag += tag;
    emptyTag += L"/>";

    start = xml.find(emptyTag, pos); 
    if (start != wstring::npos) {
        start += emptyTag.length();
        end = start;
        return 0;  // found!
    }

    //
    // Now search for the <tag>Content</tag>
    //
    openTag   = L"<";
    openTag  += tag;
    openTag  += L">";

    closeTag  = L"</";
    closeTag += tag;
    closeTag += L">";

    start = xml.find(openTag, pos); 
    if (start != wstring::npos) {
        start += openTag.length();

        end = xml.find(closeTag, start);
        if (end == wstring::npos) {
            goto not_found;
        }

        content = xml.substr(start, end-start);
    }
    
    else {
        goto not_found;
    }

    return 0;

not_found:

    return 1;
}




/**
 * Returns the path of current user's application data folder. 
 * For example: 
 * "C:\Documents And Settings\Settimio\Application Data"
 * The string returned is allocated new, so MUST be freed by caller.
 * Returns NULL in case of errors (set lastErrorMessage).
 *
 * @return : path of current user's tmp folder under 'application data'
 */
WCHAR* readAppDataPath() {

    // Get 'application data' folder for current user.
    WCHAR appDataPath[MAX_PATH_LENGTH];
    if ( FAILED(SHGetFolderPath(NULL, 
                                CSIDL_APPDATA | CSIDL_FLAG_CREATE,
                                NULL,
                                0,
                                appDataPath)) ) {
        DWORD code = GetLastError();
        char* msg = readSystemErrorMsg(code);
        sprintf (lastErrorMsg, ERR_APPDATA_PATH, code, msg);
        delete [] msg;
        return NULL;
    }

    int len = wcslen(appDataPath) + wcslen(FUNAMBOL_DIR_NAME) + wcslen(OLPLUGIN_DIR_NAME) + 3;
    WCHAR* dataPath = new WCHAR[len];
    wsprintf(dataPath, L"%s\\%s\\%s", appDataPath, FUNAMBOL_DIR_NAME, OLPLUGIN_DIR_NAME);

    return dataPath;
}


/**
 * Returns the path of file where data files for current user are stored. 
 * It is located under 'application data' folder. For example: 
 * "C:\Documents And Settings\Settimio\Application Data\Funambol\Outlook Plug-in"
 * The string returned is allocated new, so MUST be freed by caller.
 * Returns NULL in case of errors (set lastErrorMessage).
 *
 * @return : path of current user's tmp folder under 'application data'
 */
WCHAR* readDataPath(const WCHAR* itemType) {
    
    WCHAR* dataPath = readAppDataPath();
    if (!dataPath || !itemType) {
        return NULL;
    }

    WCHAR* oldItemsPath = new WCHAR[wcslen(dataPath) + wcslen(itemType) + 5];
    wsprintf(oldItemsPath, L"%s\\%s.db", dataPath, itemType);

    if (dataPath) {
        delete [] dataPath;
    }
    return oldItemsPath;
}



/**
 * Reads a file from filesystem and write its content into a string.
 * @param filePath : the path of file to read
 * @return         : the wstring with content read (empty if file not found)
 */
wstring readFromFile(const wstring& filePath) {
    
    FILE* f;
    if ((f  = _wfopen(filePath.c_str(), L"r")) == NULL) {
        // File does not exists...
        return EMPTY_WSTRING;
    }

    wstring content = EMPTY_WSTRING;
	WCHAR line[1024];
    while(fgetws(line, 1024, f) != NULL) {
        content.append(line);
    }

    fflush(f);
    fclose(f);

    return content;
}

/**
 * Write the passed string 'content' into a file on filesystem (wstring version).
 * If file not found, it is created.
 * @param content  : the string to write
 * @param filePath : path of file to write
 * @param mode     : [OPTIONAL] mode of writing (default = "w" = writing, content destroyed)
 * @return         : 0 if no errors
 */
int writeToFile(const wstring& content, const wstring& filePath, const WCHAR* mode) {

    if (!mode) {
        sprintf(lastErrorMsg, ERR_FILE_OPEN_MODE);
        return 1;
    }
    
    FILE* f;
    if ((f  = _wfopen(filePath.c_str(), mode)) == NULL) {
        sprintf(lastErrorMsg, ERR_WFILE_OPEN, filePath.c_str());
        return 1;
    }

    if (fwprintf(f, content.c_str()) < 0) {
        sprintf(lastErrorMsg, ERR_WFILE_WRITE, filePath.c_str());
        return 1;
    }

    fclose(f);
    return 0;
}


/**
 * Write the passed string 'content' into a file on filesystem (string version).
 * If file not found, it is created.
 * @param content  : the string to write
 * @param filePath : path of file to write
 * @param mode     : [OPTIONAL] mode of writing (default = "w" = writing, content destroyed)
 * @return         : 0 if no errors
 */
int writeToFile(const string& content, const string& filePath, const char* mode) {
    
    if (!mode) {
        sprintf(lastErrorMsg, ERR_FILE_OPEN_MODE);
        return 1;
    }

    FILE* f;
    if ((f  = fopen(filePath.c_str(), mode)) == NULL) {
        sprintf(lastErrorMsg, ERR_FILE_OPEN, filePath.c_str());
        return 1;
    }

    if (fprintf(f, content.c_str()) < 0) {
        sprintf(lastErrorMsg, ERR_FILE_WRITE, filePath.c_str());
        return 1;
    }

    fclose(f);
    return 0;
}


/**
 * Create directories for data files under 'application data' dir.
 * @return : 0 if no errors
 */
int makeDataDirs() {
    int err = 0;

    // Get 'application data' folder for current user.
    WCHAR appDataPath[MAX_PATH_LENGTH];
    if ( FAILED(SHGetFolderPath(NULL, 
                                CSIDL_APPDATA | CSIDL_FLAG_CREATE,
                                NULL,
                                0,
                                appDataPath)) ) {
        DWORD code = GetLastError();
        char* msg = readSystemErrorMsg(code);
        sprintf (lastErrorMsg, ERR_APPDATA_PATH, code, msg);
        delete [] msg;
        return 1;
    }

    // Create 'Funambol' directory
    wstring tmpPath = appDataPath;
    tmpPath += L"\\";
    tmpPath += FUNAMBOL_DIR_NAME;
    if ( _wmkdir(tmpPath.c_str()) ) {
        _get_errno(&err);
        if (err != EEXIST) {
            sprintf(lastErrorMsg, ERR_DIR_CREATE, tmpPath.c_str());
            return 1;
        }
    }

    // Create 'Outlook Plug-in' directory
    tmpPath += L"\\";
    tmpPath += OLPLUGIN_DIR_NAME;
    if ( _wmkdir(tmpPath.c_str()) ) {
        _get_errno(&err);
        if (err != EEXIST) {
            sprintf(lastErrorMsg, ERR_DIR_CREATE, tmpPath.c_str());
            return 1;
        }
    }

    return 0;
}



/**
 * Returns the syncMutex unique name: "fol-syncInProgress-xxxxxxx".
 * "xxxxxxx" is the BeginSync timestamp, stored in win registry (HKCU).
 * We don't use OutlookConfig object because the mutex name can be
 * asked BEFORE the initialization of client (so config not yet available).
 * So here we directly access to the win registry.
 */
string getSyncMutexName() {

    DMTree* dmt          = NULL;
    ManagementNode* node = NULL;
    char* value          = NULL;
    string ret = SYNC_MUTEX_NAME;      // By default.

    char context[DIM_MANAGEMENT_PATH];
    sprintf(context, "%s%s%s", APPLICATION_URI, CONTEXT_SPDS_SYNCML, CONTEXT_EXT);

    // Get value.
    dmt = DMTreeFactory::getDMTree(context);
    if (!dmt)   goto finally;
    node = dmt->readManagementNode(context);
    if (!node)  goto finally;
    value = node->readPropertyValue(PROPERTY_SYNC_BEGIN);
    if (!value) goto finally;

    ret.append("-");
    ret.append(value);

finally:
    if (dmt)   delete dmt;
    if (node)  delete node;
    if (value) delete [] value;

    return ret;
}



/**
 * Utility function to retrieve the correspondant message for a generic System error.
 * Pointer returned is allocated new, must be freed by caller.
 *
 * @param errorCode : the code of error (obtained by 'getLastError()')
 * @return          : the (new allocated) error message
 */
char* readSystemErrorMsg(DWORD errorCode) {

    if (!errorCode) {
        errorCode = GetLastError();
    }
    
    char* errorMessage = new char[512];
    memset(errorMessage, 0, 512);

    FormatMessageA(
                FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS,
                NULL,
                errorCode,
                MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
                errorMessage, 
                512, 
                NULL);

    if (!errorMessage || strlen(errorMessage) == 0) {
        sprintf(errorMessage, ERR_UNKNOWN);
    }
    return errorMessage;
}


/**
 * Function used to display a message-box with a message for the user.
 * The message-box is displayed only if this is NOT a scheduled sync
 * (otherwise the message will be redirected to LOG file)
 * 
 * @param message : the message to display
 * @param title   : [OPTIONAL] the title of the message box (default = MSGBOX_ERROR_TITLE)
 * @param flags   : [OPTIONAL] flags for the message-box    (default = MB_OK | MB_ICONEXCLAMATION)
 * @return        : the value returned from MessageBox call
 */
int safeMessageBox(const char* message, const char* title, unsigned int flags) {

    int ret = 0;
    if (!message) {
        return -1;
    }

    OutlookConfig* config = OutlookConfig::getInstance();

    // Normal sync
    if (config->getScheduledSync() == false) {
        WCHAR* wtitle = NULL;
        WCHAR* wmessage = toWideChar(message);

        if (!flags) {
            flags = MB_OK | MB_ICONEXCLAMATION | MB_SETFOREGROUND | MB_APPLMODAL;
        }

        if (!title) {
            wtitle = wstrdup(WMSGBOX_ERROR_TITLE);
        }
        else {
            wtitle = toWideChar(title);
        }

        ret = MessageBox(NULL, wmessage, wtitle, flags);

        if (wmessage) delete [] wmessage;
        if (wtitle)   delete [] wtitle;
    }

    // Scheduled sync
    else {
        LOG.error(message);
    }

    return ret;
}


/**
 * Function used to display a message-box with a message for the user. (WCHAR version)
 * The message-box is displayed only if this is NOT a scheduled sync
 * (otherwise the message will be redirected to LOG file)
 * 
 * @param message : the message to display
 * @param title   : [OPTIONAL] the title of the message box (default = MSGBOX_ERROR_TITLE)
 * @param flags   : [OPTIONAL] flags for the message-box    (default = MB_OK | MB_ICONEXCLAMATION)
 * @return        : the value returned from MessageBox call
 */
int wsafeMessageBox(const WCHAR* wmessage, const WCHAR* wtitle, unsigned int flags) {

    int ret = 0;
    if (!wmessage) {
        return -1;
    }

    OutlookConfig* config = OutlookConfig::getInstance();

    // Normal sync
    if (config->getScheduledSync() == false) {

        if (!flags) {
            flags = MB_OK | MB_ICONEXCLAMATION | MB_SETFOREGROUND | MB_APPLMODAL;
        }
        if (!wtitle) {
            wtitle = WMSGBOX_ERROR_TITLE;
        }

        ret = MessageBox(NULL, wmessage, wtitle, flags);
    }

    // Scheduled sync
    else {
        char* message = toMultibyte(wmessage);
        LOG.error(message);
        delete [] message;
    }

    return ret;
}


/**
 * Prints a message (char*) into LOG file. Used by UI.
 */
void printLog(const char* msg, const char* level) {

    if (!msg || !strlen(msg)) {
        return;
    }

    if (!strcmp(level, LOG_INFO)) {
        LOG.info("[UI] %s", msg);
    }
    else if (!strcmp(level, LOG_DEBUG)) {
        LOG.debug("[UI] %s", msg);
    }
    else if (!strcmp(level, LOG_ERROR)) {
        LOG.error("[UI] %s", msg);
    }
    return;
}

/**
 * Prints a message (WCHAR*) into LOG file. Used by UI.
 */
void printLog(const WCHAR* wmsg, const char* level) {

    if (!wmsg || !wcslen(wmsg)) {
        return;
    }
    char* msg = toMultibyte(wmsg);
    printLog(msg, level);
    delete [] msg;
    return;
}



/**
 * Returns a smart name of the item passed, used for log/debug or in
 * case of errors. Usually the "Subject" is returned (if present).
 */
wstring getSafeItemName(ClientItem* cItem) {
    if (!cItem) {
        return L"(null)";
    }

    wstring name = EMPTY_WSTRING;
    try {
        name = cItem->getProperty(L"Subject");
        if (name != EMPTY_WSTRING) {
            return name;
        }
        name = cItem->getID();
        if (name != EMPTY_WSTRING) {
            wstring ret = L"ID = ";
            ret += name;
            return ret;
        }
    }
    catch (ClientException* e) {
        LOG.debug(DBG_SAFE_ITEM_NAME, e->getErrorMsg()); 
    }
    return L"(new item)";
}




// Used by UI
int syncSourceNameToIndex(const char* syncSource){
    int value = 0;

    if(! strcmp(syncSource,"appointment")) {
        value = SYNCSOURCE_CALENDAR;
    }
    else if(! strcmp(syncSource,"contact")) {
        value = SYNCSOURCE_CONTACTS;
    }
    else if(! strcmp(syncSource,"task")) {
        value = SYNCSOURCE_TASKS;
    }
    else if(! strcmp(syncSource,"note")) {
        value = SYNCSOURCE_NOTES;
    }
    return value;
}


/**
 * Prints a smart synchronization report table into LOG file.
 */
void printReport(SyncReport* sr, WindowsSyncSource** sources) {

    if (sr == NULL) {
        LOG.debug("SyncReport is NULL");
        return;
    }

    // Just to remove any "\r\n" at the end of system error messages.
    int i = strlen(lastErrorMsg) - 1;
    if (i>0) {
        while (lastErrorMsg[i] == '\n' || lastErrorMsg[i] == '\r') {
            if (i==0) break;
            lastErrorMsg[i] = 0;
           i--;
        }
    }

    // Update SyncReport with last error from sync
    sr->setLastErrorCode(lastErrorCode);
    sr->setLastErrorMsg(lastErrorMsg);

    SyncSourceReport* ssr = NULL;
    string res;
    char tmp[1024];

    res =      "============================================================\n";
    res.append("================   SYNCHRONIZATION REPORT   ================\n");
    res.append("============================================================\n\n");

    if (sr->getLastErrorCode() == 0) {
        res.append("SYNCHRONIZATION COMPLETED SUCCESSFULLY!\n");
        res.append("---------------------------------------\n\n");
    }
    else {
        res.append("SYNCHRONIZATION COMPLETED WITH ERRORS\n");
        res.append("-------------------------------------\n");
        sprintf(tmp, "Last error message = \"%s\"\n", sr->getLastErrorMsg());
        res.append(tmp);
        sprintf(tmp, "Last error code    = %d\n\n", sr->getLastErrorCode());
        res.append(tmp);
    }

    i=0;
    while (sources[i]) {
        ssr = sr->getSyncSourceReport(sources[i]->getConfig().getName());
        if (!ssr) {
            i++;
            continue;
        }

        // SourceName
        char* name = friendlyName(ssr->getSourceName());
        sprintf(tmp, "%s:\n", name);
        res.append(tmp);
        for (unsigned int j=0; j<strlen(name)+1; j++) {
            res.append("-");
        }
        res.append("\n");

        //
        // Calculate #items for sources table
        //
        int newCOk  = ssr->getItemReportSuccessfulCount(CLIENT, COMMAND_ADD);
        int newCTot = ssr->getItemReportCount          (CLIENT, COMMAND_ADD);
        int newSOk  = ssr->getItemReportSuccessfulCount(SERVER, COMMAND_ADD);
        int newSTot = ssr->getItemReportCount          (SERVER, COMMAND_ADD);
        int modCOk  = ssr->getItemReportSuccessfulCount(CLIENT, COMMAND_REPLACE);
        int modCTot = ssr->getItemReportCount          (CLIENT, COMMAND_REPLACE);
        int modSOk  = ssr->getItemReportSuccessfulCount(SERVER, COMMAND_REPLACE);
        int modSTot = ssr->getItemReportCount          (SERVER, COMMAND_REPLACE);
        int delCOk  = ssr->getItemReportSuccessfulCount(CLIENT, COMMAND_DELETE);
        int delCTot = ssr->getItemReportCount          (CLIENT, COMMAND_DELETE);
        int delSOk  = ssr->getItemReportSuccessfulCount(SERVER, COMMAND_DELETE);
        int delSTot = ssr->getItemReportCount          (SERVER, COMMAND_DELETE);

        // Subtract the #of item with status 418 from total and ok (only server).
        int newSEx  = ssr->getItemReportAlreadyExistCount(SERVER, COMMAND_ADD);
        if (newSEx > 0) {
            newSTot -= newSEx;
            newSOk  -= newSEx;
        }
        int modSEx  = ssr->getItemReportAlreadyExistCount(SERVER, COMMAND_REPLACE);
        if (modSEx > 0) {
            modSTot -= modSEx;
            modSOk  -= modSEx;
        }

        // Don't show items removed internally.
        SyncMode mode = sources[i]->getSyncMode();
        if (mode == SYNC_REFRESH_FROM_SERVER) {
            delCOk  = 0;
            delCTot = 0;
        }

        //
        // Source ok
        //
        if (ssr->checkState()) {

            // Source not synced
            if (sources[i]->getConfig().getIsSynced() == false) {
                res.append("    Not synced.\n\n");
                i++;
                continue;
            }

            // Some items dropped on client
            if ((newCTot-newCOk > 0) || (modCTot-modCOk > 0)) {
                res.append("    Sync completed - WARNING: some items have not been inserted on Outlook.\n");
                if (ssr->getLastErrorCode()) {
                    sprintf(tmp, "    Last error = \"%s\" (code %d)\n", ssr->getLastErrorMsg(), ssr->getLastErrorCode());
                    res.append(tmp);
                }
                if (!sr->getLastErrorCode()) {
                    // This will prompt a warning msg on the UI.
                    lastErrorCode = ERR_CODE_DROPPED_ITEMS;
                    sr->setLastErrorCode(ERR_CODE_DROPPED_ITEMS);
                }
            }
            // Some items dropped on server
            else if ((newSTot-newSOk > 0) || (modSTot-modSOk > 0)) {
                res.append("    Sync completed - WARNING: some items have not been inserted on server.\n");
                if (ssr->getLastErrorCode()) {
                    sprintf(tmp, "    Last error = \"%s\" (code %d)\n", ssr->getLastErrorMsg(), ssr->getLastErrorCode());
                    res.append(tmp);
                }
            }
            // Everything OK!
            else {
                res.append("    Sync completed successfully!\n");
            }


            char* usedMode = syncModeName(mode);
            const char* origMode = sources[i]->getConfig().getSync();
            sprintf(tmp, "    Sync type: %s", usedMode);
            res.append(tmp);

            // Slow requested by server
            if (strcmp(usedMode, origMode)) {
                sprintf(tmp, " (requested by Server)", usedMode);
                res.append(tmp);
            }
            res.append("\n\n");
        }

        //
        // Source error
        //
        else {
            sprintf(tmp, "    Sync failed: %s (code = %d)\n\n", ssr->getLastErrorMsg(), ssr->getLastErrorCode());
            res.append(tmp);
        }

        //
        // Format NEW-MOD-DEL table
        //
        res.append("            | on Client | on Server\n");
        res.append("    --------|-----------|----------\n");
        sprintf(tmp, "    New     | %4d/%4d | %4d/%4d ", newCOk, newCTot, newSOk, newSTot);
        res.append(tmp);
        if (newSEx) {
            sprintf(tmp, " (%d already exist on Server)", newSEx);
            res.append(tmp);
        }
        res.append("\n");
        sprintf(tmp, "    Updated | %4d/%4d | %4d/%4d ", modCOk, modCTot, modSOk, modSTot);
        res.append(tmp);
        if (modSEx) {
            sprintf(tmp, " (%d already exist on Server)", modSEx);
            res.append(tmp);
        }
        res.append("\n");
        sprintf(tmp, "    Deleted | %4d/%4d | %4d/%4d \n\n\n", delCOk, delCTot, delSOk, delSTot);
        res.append(tmp);

        i++;
    }

    LOG.info("\n%s", res.c_str());
}


/**
 * Returns a friendly name for the passed syncsource name.
 */
char* friendlyName(const char* sourceName) {

    if (!sourceName) {
        return EMPTY_STRING;
    }
    if (!strcmp(sourceName, APPOINTMENT_)) {
        return "Calendar";
    }
    else if (!strcmp(sourceName, CONTACT_)) {
        return "Contacts";
    }
    else if (!strcmp(sourceName, NOTE_)) {
        return "Notes";
    }
    else if (!strcmp(sourceName, TASK_)) {
        return "Tasks";
    }
    return EMPTY_STRING;
}


/**
 * Used to encrypt private data (username/password) using 'B64(DES(data))'.
 * @param data   the data to encrypt
 * @return       the data encrypted (new buffer: must be deleted by caller)
 */
char* encryptData(const char* data) {
    
    if (!data) return NULL;
    char*   desData = NULL;
    char*   b64Data = NULL;
    char* data1 = (char*)data;

    //
    // 1. DES encryption
    //
    DataTransformer* dt = new DESEncoder();
    TransformationInfo info;
    if (dt == NULL) {
        goto error;
    }
    info.size = strlen(data1);
    info.password = "SettimioSevero";
    desData = dt->transform(data1, info);
    if (!desData || info.newReturnedData == FALSE) {
        goto error;
    }
    if (dt) delete dt;

    //
    // 2. B64 encoding
    //
    dt = new B64Encoder();
    // info.size is already correct :)
    info.password = NULL;
    b64Data = dt->transform(desData, info);
    goto exit;


error:
    LOG.error(ERR_ENCRYPT_DATA);
    b64Data = stringdup(data);

exit:
    if (desData) delete [] desData;
    if (dt)      delete dt;
    return b64Data;
}



/**
 * Used to decrypt private data (username/password) using 'dec_DES(dec_B64(b64Data))'.
 * @param b64Data   the data to decrypt
 * @return          the data decrypted (new buffer: must be deleted by caller)
 */
char* decryptData(const char* b64Data) {
    
    if (!b64Data) return NULL;
    char* desData = NULL;
    char* data = NULL;

    // MUST copy, this buffer will be modified and 
    // it's the one returned from this function.
    char* b64Data1 = stringdup(b64Data);

    //
    // 1. B64 decoding
    //
    DataTransformer* dt = new B64Decoder();
    TransformationInfo info;
    if (dt == NULL) {
        goto error;
    }
    info.size = strlen(b64Data1);
    info.password = NULL;
    desData = dt->transform(b64Data1, info);
    if (dt) delete dt;


    //
    // 2. DES decryption
    //
    dt = new DESDecoder();
    // info.size is already correct :)
    info.password = "SettimioSevero";
    data = dt->transform(desData, info);
    if (!data) {
        goto error;
    }
    data[info.size] = 0;
    goto exit;

error:
    LOG.error(ERR_DECRYPT_DATA);
    data = stringdup(EMPTY_STRING);

exit:
    // NO! buffer used is the same passed to both decoder!!
    // if (desData) delete [] desData;
    // if (b64Data1) delete [] b64Data1;
    if (dt)       delete dt;
    return data;
}


/**
 * Returns an integer value rapresenting the build number read from
 * the string version passed (e.g. "6.1.12" -> 600112). 
 */
int getBuildNumberFromVersion(const char* swv) {

    int major=0, minor=0, build=0;
    if (!swv) {
        return 0;
    }
    sscanf(swv, "%d.%d.%d", &major, &minor, &build);
    return (major*10000 + minor*100 + build);
}


//
// ----------------------------------- Time/date functions ----------------------------------------
//

/**
 * Variant time (double format)  ->  Timestamp (long format).
 * NOTE: the internal call 'mktime()' adjusts time to UTC... this is OK because
 *       timestamps always refer to UTC time.
 *
 * @param vTime : input time in variant format (milliseconds from Jan 1 1900)
 * @return      : output time as timestamp (seconds from Jan 1 1970)
 */
long variantTimeToTimeStamp(const double vTime) {

    tm t;
    SYSTEMTIME st;
    VariantTimeToSystemTime(vTime, &st);

    t.tm_hour = st.wHour;
    t.tm_mday = st.wDay;
    t.tm_min  = st.wMinute;
    t.tm_mon  = st.wMonth - 1;      // range [0-11]
    t.tm_sec  = st.wSecond;
    t.tm_wday = st.wDayOfWeek;      // range [0-6] (0 = sunday)
    t.tm_year = st.wYear - 1900;    // From 1900
    t.tm_isdst= -1;                 // less than 0 means to auto-adjust standard time or daylight savings time.

    // This call automatically converts to UTC!
    return (long)mktime(&t);
}



/**
 * Returns TRUE if startdate = 00:00 and enddate = 23:59.
 * This is an all-day-event for VCal format.
 */
BOOL isAllDayEventInterval(const DATE startdate, const DATE enddate) {

    SYSTEMTIME ststart, stend;
    VariantTimeToSystemTime(startdate, &ststart);
    VariantTimeToSystemTime(enddate, &stend);
    
    BOOL ret = FALSE;
    if (ststart.wHour == 0 && ststart.wMinute == 0 &&
        stend.wHour == 23 && stend.wMinute == 59) {
        ret = TRUE;
    }

    return ret;
}


/**
 * Converts a date in double format into "yyyyMMdd" format.
 * All-day-event items need this date format for VCal.
*/
void doubleToSystemTimeDateForVCal(wstring& systemDate, const DATE doubleDate) {

    SYSTEMTIME t;
    VariantTimeToSystemTime(doubleDate, &t);

    WCHAR month [10];
    WCHAR day   [10];
    WCHAR date  [10];

    // month
    if (t.wMonth < 10)
        wsprintf(month, TEXT("0%i"), t.wMonth);
    else
        wsprintf(month, TEXT("%i"), t.wMonth);

    // day
    if (t.wDay  < 10)
        wsprintf(day, TEXT("0%i"), t.wDay);
    else
        wsprintf(day, TEXT("%i"), t.wDay);

    wsprintf(date, TEXT("%i%s%s"), t.wYear, month, day);
    systemDate = date;
}



/**
  * Converts recurrence pattern extracted from ClientItem passed to an RRULE (for VCal).
  * Returns a string containing the RRULE for VCal/ICal.
  *
  * @param cRec : the ClientRecurrence object to read from
  * @return     : the RRULE formatted string
  */
wstring clientRecurrenceToRRULE(ClientRecurrence* cRec) {

    if (!cRec) {
        return EMPTY_WSTRING;
    }
    wstring ret;
    WCHAR* days = NULL;
    WCHAR recurrence[100];
    wsprintf(recurrence, EMPTY_WSTRING);


    //
    // Get all rec properties from ClientRecurrence
    //
    const int recType     = cRec->getRecurrenceType  ();
    const wstring pEnd    = cRec->getPatternEndDate  ();
    const wstring pStart  = cRec->getPatternStartDate();
    const BOOL noEnd      = cRec->getNoEndDate       ();
    const int interval    = cRec->getInterval        ();
    const int occurences  = cRec->getOccurrences     ();
    const int dayofweek   = cRec->getDayOfWeekMask   ();
    const int dayofmonth  = cRec->getDayOfMonth      ();
    const int instance    = cRec->getInstance        ();
    const int monthofyear = cRec->getMonthOfYear     ();


    //
    // Format the RRULE
    //
    switch(recType) {

        // Daily = 0
        case olRecursDaily: {
            //if(noEnd)
            //    occurences = 0;  //occurences should be already = 0 to indicate event that does not end
            if(interval >0)
                wsprintf(recurrence, TEXT("D%ld #%ld"), interval, occurences);
            else if (dayofweek > 0) {
                WCHAR* days = extractDayOfWeek(dayofweek);
                if (days) {
                    wsprintf(recurrence, TEXT("W1 %s %s #%ld"), days, pEnd.c_str(), occurences);
                    delete [] days;
                }
            }
            break;
        }

        // Weekly = 1
        case olRecursWeekly: {
            if (dayofweek > 0) {
                days = extractDayOfWeek(dayofweek);
                if (days) {
                    if(noEnd)
                        wsprintf(recurrence, TEXT("W%ld %s #0"), interval, days);
                    else
                        wsprintf(recurrence, TEXT("W%ld %s%s #%ld"), interval, days, pEnd.c_str(), occurences);
                    delete [] days;
                }
            }
            break;
        }
        
        // Monthly = 2
        case olRecursMonthly: {
            if(dayofmonth > 0) {
                if(noEnd)
                    wsprintf(recurrence, TEXT("MD%ld %ld #0"), interval, dayofmonth);
                else {
                    wsprintf(recurrence, TEXT("MD%ld %ld %s #%ld"), interval, dayofmonth, pEnd.c_str(), occurences);
                }
                break;
            }
        }
        
        // MonthNth = 3
        case olRecursMonthNth: {
            if(instance>0 && dayofweek>0) {
                days = extractDayOfWeek(dayofweek);
                if(days) {
                    if(noEnd)
                        wsprintf(recurrence, TEXT("MP%ld %ld+ %s #0"), interval, instance, days);
                    else
                        wsprintf(recurrence, TEXT("MP%ld %ld+ %s%s #%ld"), interval, instance, days, pEnd.c_str(), occurences);
                    delete [] days;
                }
            }
            break;
        }

        // Yearly = 5
        case olRecursYearly: {
            if(dayofmonth>0 && monthofyear>0) {
                if(noEnd)
                    wsprintf(recurrence, TEXT("YM1 %ld #0"), monthofyear);
                else
                    wsprintf(recurrence, TEXT("YM1 %ld #%ld"), monthofyear, occurences);
            }
            break;
        }
        
        // YearNth = 6
        case olRecursYearNth: {
            if(dayofweek>0 && instance>0) {
                days = extractDayOfWeek(dayofweek);
                if(days) {
                    if(noEnd)
                        wsprintf(recurrence, TEXT("MP12 %ld+ %s #0"), instance, days);
                    else
                        wsprintf(recurrence, TEXT("MP12 %ld+ %s%s #%ld"), instance, days, pEnd.c_str(), occurences);
                    delete [] days;
                }
            }
            break;
        }
    }

    ret = recurrence;
    return ret;
}


/**
  * Converts an RRULE (VCal) into a recurrence pattern, and 
  * fills the passed ClientRecurrence object.
  *
  * @param rrule     : the RRULE formatted string to read from
  * @param cRec      : the ClientRecurrence object to fill with rec properties
  * @param startDate : the start date of event/task (necessary for ClientRecurrence)
  * @return          : true if RRULE valid and parsed correctly.
  */
bool rRULEToClientRecurrence(const wstring& rrule, ClientRecurrence* cRec, const DATE startDate) {

    bool ret = false;
    WCHAR* str = wstrdup(rrule.c_str());
    WCHAR seps[] = TEXT(" ");
    WCHAR* token = wcstok(str, seps);

    int recType     = -1;
    int interval    =  0;
    int occurences  =  0;
    int dayOfMonth  = -1;
    int weekOfMonth = -1;
    int monthOfYear = -1;


    //
    // First token will be: "D|W|MP|MD|YM|YD<interval>"
    //
    if(token[0] == TEXT('D')) {
        recType = olRecursDaily;
        token ++;
    }
    else if(token[0] == TEXT('W')) {
        recType = olRecursWeekly;
        token ++;
    }
    else if(token[0] == TEXT('M') && token[1] == TEXT('D')) {
        recType = olRecursMonthly;
        token += 2;
    }
    else if(token[0] == TEXT('M') && token[1] == TEXT('P')) {
        recType = olRecursMonthNth;
        token += 2;
    }
    else if(token[0] == TEXT('Y') && token[1] == TEXT('D')) {
        // "YD" Not supported!!
        goto error;
    }
    else if(token[0] == TEXT('Y') && token[1] == TEXT('M')) {
        recType = olRecursYearly;
        token += 2;
    }

    interval = _wtoi(token);
    if(!interval || recType == -1) {
        goto error;
    }


    WCHAR* days = new WCHAR[30];
    wcscpy(days, EMPTY_WSTRING);

    WCHAR* mOfYear = new WCHAR[20];
    wcscpy(mOfYear, EMPTY_WSTRING);


    token = wcstok(NULL, seps);
    while (token) {

        //
        // Daily = 0
        //
        if(recType == olRecursDaily) {
            if(wcschr(token, TEXT('#'))) {
                cRec->setRecurrenceType(recType);
                cRec->setInterval(interval);
                occurences = _wtoi(token+1);
                if(occurences == 0)             cRec->setNoEndDate(TRUE);
                else                            cRec->setOccurrences(occurences);
                token ++;
            }
            else if(token[8] == TEXT('T')) {
                cRec->setRecurrenceType(recType);
                cRec->setPatternEndDate(token);
                cRec->setInterval(interval);
            }
        }

        //
        // Weekly = 1
        //
        else if(recType == olRecursWeekly) {
            if(wcschr(token, TEXT('#'))) {
                cRec->setRecurrenceType(recType);
                cRec->setInterval(interval);
                occurences = _wtoi(token+1);
                if(occurences == 0)             cRec->setNoEndDate(TRUE);
                else                            cRec->setOccurrences(occurences);
                if(wcscmp(days, EMPTY_WSTRING)) cRec->setDayOfWeekMask(calculateDayOfWeek(days));
                else                            cRec->setDayOfWeekMask(getWeekDay(startDate));
                token++;
            }
            else if(token[8] == TEXT('T')) {
                cRec->setRecurrenceType(recType);
                cRec->setPatternEndDate(token);
                cRec->setInterval(interval);
                if(wcscmp(days,EMPTY_WSTRING)) cRec->setDayOfWeekMask(calculateDayOfWeek(days));
                else                           cRec->setDayOfWeekMask(getWeekDay(startDate));
            }
            else if(isDayOfWeek(token)) {
                wcscat(days, token);
                wcscat(days, TEXT(" "));
            }
        }

        //
        // Monthly = 2
        //
        else if(recType == olRecursMonthly) {
            // VCalendar supports as Monthly by day rules: MD<interval> listofdates|listofdates(from end of month) <end tag>
            // On outlook recurrence pattern only MD<interval> day <end tag> can be represented
            if(wcschr(token, TEXT('#'))) {
                cRec->setRecurrenceType(recType);
                cRec->setDayOfMonth(dayOfMonth);
                cRec->setInterval(interval);
                occurences = _wtoi(token+1);
                if(occurences == 0)             cRec->setNoEndDate(TRUE);
                else                            cRec->setOccurrences(occurences);
                token++;
            }
            else if(token[8] == TEXT('T')) {
                cRec->setRecurrenceType(recType);
                cRec->setPatternEndDate(token);
                cRec->setInterval(interval);
                cRec->setDayOfMonth(dayOfMonth);
            }
            else {
                if(dayOfMonth != -1) {
                    goto error;
                }
                dayOfMonth = _wtoi(token);
                if(dayOfMonth == 0) {
                    goto error;
                }
            }
        }

        //
        // MonthNth = 3
        //
        else if(recType == olRecursMonthNth) {
            if(wcschr(token, TEXT('#'))) {
                cRec->setRecurrenceType(recType);
                cRec->setInterval(interval);
                cRec->setInstance(weekOfMonth);
                cRec->setDayOfWeekMask(calculateDayOfWeek(days));
                occurences = _wtoi(token+1);
                if(occurences == 0)             cRec->setNoEndDate(TRUE);
                else                            cRec->setOccurrences(occurences);
                token++;
            }
            else if(token[8] == TEXT('T')) {
                cRec->setRecurrenceType(recType);
                cRec->setPatternEndDate(token);
                cRec->setInterval(interval);
                cRec->setInstance(weekOfMonth);
                cRec->setDayOfWeekMask(calculateDayOfWeek(days));
            }
            else if(isDayOfWeek(token)) {
                wcscat(days, token);
                wcscat(days, TEXT(" "));
            }
            else {
                if(token[1] != TEXT('+') && token[1] != TEXT('-')) {
                    goto error;
                }
                WCHAR sWeek[] = TEXT("\0\0");
                sWeek[0] = token[0];
                weekOfMonth = _wtoi(sWeek);
                if(token[1] == TEXT('-')) {
                    // it's the # of weeks to the end of month...
                    weekOfMonth = 5 - weekOfMonth;
                }
                if(weekOfMonth > 5 || weekOfMonth < 1) {
                    goto error;
                }
            }
        }

        //
        // Yearly = 5
        //
        else if(recType == olRecursYearly) {
            // Expected sequence will be YM1 month <end tag>
            if(wcschr(token, TEXT('#'))) {
                cRec->setRecurrenceType(recType);
                cRec->setInterval      (interval);
                cRec->setMonthOfYear   (monthOfYear);
                cRec->setDayOfMonth    (dayOfMonth);
                occurences = _wtoi(token+1);
                if(occurences == 0)             cRec->setNoEndDate(TRUE);
                else                            cRec->setOccurrences(occurences);
                token++;
            }
            else if(token[8] == TEXT('T')) {
                cRec->setRecurrenceType(recType);
                cRec->setPatternEndDate(token);
                cRec->setInterval      (interval);
                cRec->setMonthOfYear   (monthOfYear);
                cRec->setDayOfMonth    (dayOfMonth);
            }
            else {
                if(wcscmp(mOfYear, EMPTY_WSTRING)) {
                    goto error;
                }
                wcscat(mOfYear, token);
                monthOfYear = _wtoi(mOfYear);
                SYSTEMTIME st;
                VariantTimeToSystemTime(startDate, &st);
                dayOfMonth = st.wDay;
                cRec->setRecurrenceType(recType);
                cRec->setInterval      (interval);
                cRec->setMonthOfYear   (monthOfYear);
                cRec->setDayOfMonth    (dayOfMonth);
            }
        }
        token = wcstok(NULL, seps);
    }

    ret = true;
    goto finally;

error:
    LOG.debug(DBG_RRULE_BAD_FORMAT, rrule.c_str());
    ret = false;
    goto finally;

finally:
    if (str)     delete [] str;
    if (days)    delete [] days;
    if (mOfYear) delete [] mOfYear;
    return ret;
}




// --------- Days of week MASK -----------
//enum OlDaysOfWeek {
//    olSunday    = 1,
//    olMonday    = 2,
//    olTuesday   = 4,
//    olWednesday = 8,
//    olThursday  = 16,
//    olFriday    = 32,
//    olSaturday  = 64,
//};
//----------------------------------------
/**
 * daysOfWeekMask -> string
 * Returns a string with days of week formatted (like "SU MO TU FR"),
 * based on the daysOfWeek mask passed.
 * Returns a new allocated string, must be freed by the caller.
 */
WCHAR* extractDayOfWeek(long l) {

    if (l<0 || l>128)
        return NULL;

    //SU MO TU WE TH FR SA
    WCHAR* ret = new WCHAR[22];
    wcscpy(ret, TEXT(""));

    if(l & olSunday)
        wcscat(ret, TEXT("SU "));
    if(l & olMonday)
        wcscat(ret, TEXT("MO "));
    if(l & olTuesday)
        wcscat(ret, TEXT("TU "));
    if(l & olWednesday)
        wcscat(ret, TEXT("WE "));
    if(l & olThursday)
        wcscat(ret, TEXT("TH "));
    if(l & olFriday)
        wcscat(ret, TEXT("FR "));
    if(l & olSaturday)
        wcscat(ret, TEXT("SA "));

    return ret;
}

/**
 * string -> dayOfWeekMask
 * Calculate the dayOfWeekMask based on the input string
 * of days (like "SU MO TU FR").
 */
int calculateDayOfWeek(WCHAR* in) {
    int ret = 0;

    WCHAR* index;
    index = NULL;
    index = wcsstr(in, TEXT("SU"));
    if(index)
        ret += olSunday;

    index = NULL;
    index = wcsstr(in, TEXT("MO"));
    if(index)
        ret += olMonday;

    index = NULL;
    index = wcsstr(in, TEXT("TU"));
    if(index)
        ret += olTuesday;

    index = NULL;
    index = wcsstr(in, TEXT("WE"));
    if(index)
        ret += olWednesday;

    index = NULL;
    index = wcsstr(in, TEXT("TH"));
    if(index)
        ret += olThursday;

    index = NULL;
    index = wcsstr(in, TEXT("FR"));
    if(index)
        ret += olFriday;

    index = NULL;
    index = wcsstr(in, TEXT("SA"));
    if(index)
        ret += olSaturday;

    return ret;
}


/**
 * Returns true if input string is a day of week.
 */
bool isDayOfWeek(WCHAR* data) {

    bool ret = false;
    WCHAR* weekDay[] = {TEXT("SU"), TEXT("MO"), TEXT("TU"), TEXT("WE"), TEXT("TH"), TEXT("FR"), TEXT("SA")};

    for(int i = 0; i < 7 ; i++) {
        if(!wcscmp(data, weekDay[i]))
            return true;
    }
    return ret;
}

long getWeekDay(DATE dt) {
    long ret = 0;

    SYSTEMTIME st;
    VariantTimeToSystemTime(dt, &st);

    ret = (long)(st.wDayOfWeek)*(st.wDayOfWeek);
    return ret;
}


int round(double val) {
    int v = (int)val;
    return ((val - v) > 0.5) ? v+1 : v;
}
