/*
* Copyright (C) 2003-2007 Funambol
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef INCL_CLIENTRECURRENCE
#define INCL_CLIENTRECURRENCE


#include "outlook/defs.h"
#include "outlook/ClientAppException.h"

#include <string>
#include <list>

typedef std::list<ClientAppException>     exceptionList;
typedef exceptionList::iterator           exceptionIterator;


///////////////////////////////////////////////////////////////////
class ClientRecurrence {

private:

    // Rec properties:
    int          recurrenceType;
    int          interval;
    int          monthOfYear;
    int          dayOfMonth;
    int          dayOfWeekMask;
    int          instance;
    std::wstring patternStartDate;
    BOOL         noEndDate;
    std::wstring patternEndDate;
    int          occurrences;

    std::wstring start;             // for the change-day option


    // Pointer to microsoft outlook objects.
    RecurrencePatternPtr  pRec;

    // Internal list of appointment exceptions.
    // ----------------------------------------
    // A recurrence object has a list of appointment exceptions inside it.
    // CLIENT to SERVER:
    // Exceptions are populated into list during 'read()' method, each exception can
    // be retrieved calling 'getException(index)'.
    // SERVER to CLIENT:
    // Exceptions can be appended to list calling 'addException()' method. Then all
    // exceptions will be saved all together during 'ClientAppointment::save()'.
    exceptionList appExceptions;


    // Is this rec pattern active?
    // this is linked to the property 'recurring' of Outlook event 
    // (it's set at constructor and each time setting property IsRecurring)
    bool recurring;

    // Internal use: true if values are updated with Outlook.
    // Used because all values need to be retrieved together, so this flag
    // notifies when it's necessary to update properties.
    bool isUpdated;

    // Internal use: to avoid deadlocking into safeSaveException() recursive call.
    int numRecursions;



    // TBD: verify if some props are not correct for the rec type.
    void checkIfRecIsCorrect();

    // Change-day of rec props, according on UTC <-> Local time
    bool changeDay(const std::wstring dest);


    //
    // Internal methods to manage exceptions.
    //
    // used during 'read()' method (get data from client)
    ClientAppException* getExceptionOnClient(const int index);
    const int           getExceptionsCountOnClient();
    //void              freeDestinationDays(const DATE startDate, const DATE originalDate);
    //int               safeSaveException(ClientAppException* cException);
    int                 saveException(ClientAppException* cException);


public:

    // Constructor
    ClientRecurrence();
    //ClientRecurrence(const ClientRecurrence& c);
    //ClientRecurrence operator=(const ClientRecurrence& c);

    // Destructor
    ~ClientRecurrence();


    // Set a COM pointer to this object.
    void setCOMPtr(RecurrencePatternPtr& ptr);



    /*
     * Retrieve all properties from Outlook -> set isUpdated = true.
     * If UTC is used, all props are converted to correct values.
     * 
     * @return: 0 if no errors
     */
    int read();

    /*
     * Save all the properties to Outlook.
     * Properties are verified to be consistent all together, 
     * as Outlook doesn't accept wrong values.
     * If UTC is used, all props are converted to correct values.
     * 
     * @return: 0 if no errors
     */
    int save();




    // Return true if the rec pattern is active.
    bool isRecurring();

    // Set 'recurring' to true. 
    // This is called by 'ClientAppointment.setProperty("IsRecurring", "1")'.
    void setRecurrence();

    // Reset all members, set 'recurring' to false.
    // This is called by 'ClientAppointment.setProperty("IsRecurring", "0")'.
    void clearRecurrence();




    //
    // Methods to get recurrence properties.
    // -------------------------------------
    // The first time a property is needed (isUpdated = false), all values 
    // are retrieved from Outlook calling 'read()'.
    //
    const int           getRecurrenceType  ();
    const int           getInterval        ();
    const int           getMonthOfYear     ();
    const int           getDayOfMonth      ();
    const int           getDayOfWeekMask   ();
    const int           getInstance        ();
    const std::wstring& getPatternStartDate();
    const BOOL          getNoEndDate       ();
    const std::wstring& getPatternEndDate  ();
    const int           getOccurrences     ();


    //
    // Methods to set recurrence properties.
    // -------------------------------------
    // Values passed are stored in internal members, calling 'save()' all properties 
    // are put together into Outlook rec pattern (this is for UTC transformations).
    // No test on value content is done at this level.
    // 'isUpdated' is set to false, cause setting a property the rec pattern is no
    // more updated with Outlook.
    //
    void setRecurrenceType  (const int           val);
    void setInterval        (const int           val);
    void setMonthOfYear     (const int           val);
    void setDayOfMonth      (const int           val);
    void setDayOfWeekMask   (const int           val);
    void setInstance        (const int           val);
    void setPatternStartDate(const std::wstring& val);
    void setNoEndDate       (const BOOL          val);
    void setPatternEndDate  (const std::wstring& val);
    void setOccurrences     (const int           val);

    // for the change-day option
    void setStart           (const std::wstring& val);


    //
    // To manage appointment exceptions.
    //
    ClientAppException* getException(const int index);
    const int           getExceptionsCount();
    int                 addException(ClientAppException* cException);
    void                resetExceptions();
    int                 saveAllExceptions();
    int                 removeDuplicatedExceptions();

    _AppointmentItemPtr getOccurrence(const DATE originalDate);

};

#endif
